"""
The Support Array parameters definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from xml.etree import ElementTree
from typing import Union, List, Tuple, Optional

import numpy

from sarpy.io.xml.base import Serializable, ParametersCollection, get_node_value
from sarpy.io.xml.descriptors import FloatDescriptor, StringDescriptor, StringEnumDescriptor, \
    ParametersDescriptor, SerializableListDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .utils import homogeneous_dtype


class SupportArrayCore(Serializable):
    """
    The support array base case.
    """

    _fields = ('Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS', 'NODATA')
    _required = ('Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS')
    _numeric_format = {'X0': FLOAT_FORMAT, 'Y0': FLOAT_FORMAT, 'XSS': FLOAT_FORMAT, 'YSS': FLOAT_FORMAT}
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='The support array identifier.')  # type: str
    ElementFormat = StringDescriptor(
        'ElementFormat', _required, strict=DEFAULT_STRICT,
        docstring='The data element format.')  # type: str
    X0 = FloatDescriptor(
        'X0', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: float
    Y0 = FloatDescriptor(
        'Y0', _required, strict=DEFAULT_STRICT,
        docstring='')  # type: float
    XSS = FloatDescriptor(
        'XSS', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='')  # type: float
    YSS = FloatDescriptor(
        'YSS', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='')  # type: float

    def __init__(
            self,
            Identifier: str = None,
            ElementFormat: str = None,
            X0: float = None,
            Y0: float = None,
            XSS: float = None,
            YSS: float = None,
            NODATA: str = None,
            **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        ElementFormat : str
        X0 : float
        Y0 : float
        XSS : float
        YSS : float
        NODATA : None|str
        kwargs
        """

        self._NODATA = None
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.ElementFormat = ElementFormat
        self.X0 = X0
        self.Y0 = Y0
        self.XSS = XSS
        self.YSS = YSS
        self.NODATA = NODATA
        super(SupportArrayCore, self).__init__(**kwargs)

    @property
    def NODATA(self) -> Optional[str]:
        """
        None|str: The no data hex string value.
        """

        return self._NODATA

    @NODATA.setter
    def NODATA(self, value: Optional[str]):
        if value is None:
            self._NODATA = None
            return

        if isinstance(value, ElementTree.Element):
            value = get_node_value(value)

        if isinstance(value, str):
            self._NODATA = value
        elif isinstance(value, bytes):
            self._NODATA = value.decode('utf-8')
        elif isinstance(value, int):
            raise NotImplementedError
        elif isinstance(value, float):
            raise NotImplementedError
        else:
            raise TypeError('Got unexpected type {}'.format(type(value)))

    def get_nodata_as_int(self) -> Optional[int]:
        """
        Get the no data value as an integer value.

        Returns
        -------
        None|int
        """

        if self._NODATA is None:
            return None

        raise NotImplementedError

    def get_nodata_as_float(self) -> Optional[float]:
        """
        Gets the no data value as a floating point value.

        Returns
        -------
        None|float
        """

        if self._NODATA is None:
            return None

        raise NotImplementedError

    def get_numpy_format(self) -> Tuple[numpy.dtype, int]:
        """
        Convert the element format to a numpy dtype (including endianness) and depth.

        Returns
        -------
        data: numpy.dtype
        depth: int
        """

        return homogeneous_dtype(self.ElementFormat, return_length=True)


class IAZArrayType(SupportArrayCore):
    """
    Array of scene surface heights expressed in image coordinate IAZ values (meters).
    Grid coordinates are image area coordinates (IAX, IAY).
    """

    _fields = ('Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS', 'NODATA')
    _required = ('Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS')
    # descriptors
    ElementFormat = StringEnumDescriptor(
        'ElementFormat', ('IAZ=F4;', ), _required, strict=DEFAULT_STRICT, default_value='IAZ=F4;',
        docstring='The data element format.')  # type: str

    def __init__(
            self,
            Identifier: str = None,
            ElementFormat: str = 'IAZ=F4;',
            X0: float = None,
            Y0: float = None,
            XSS: float = None,
            YSS: float = None,
            NODATA: str = None,
            **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        ElementFormat : str
        X0 : float
        Y0 : float
        XSS : float
        YSS : float
        NODATA : str
        kwargs
        """

        super(IAZArrayType, self).__init__(
            Identifier=Identifier, ElementFormat=ElementFormat, X0=X0, Y0=Y0,
            XSS=XSS, YSS=YSS, NODATA=NODATA, **kwargs)


class AntGainPhaseType(SupportArrayCore):
    """
    Antenna array with values are antenna gain and phase expressed in dB and
    cycles. Array coordinates are direction cosines with respect to the
    ACF (DCX, DCY).
    """

    _fields = ('Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS', 'NODATA')
    _required = ('Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS')
    # descriptors
    ElementFormat = StringEnumDescriptor(
        'ElementFormat', ('Gain=F4;Phase=F4;', ), _required, strict=DEFAULT_STRICT, default_value='Gain=F4;Phase=F4;',
        docstring='The data element format.')  # type: str

    def __init__(
            self,
            Identifier: str = None,
            ElementFormat: str = 'Gain=F4;Phase=F4;',
            X0: float = None,
            Y0: float = None,
            XSS: float = None,
            YSS: float = None,
            NODATA: str = None,
            **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        X0 : float
        Y0 : float
        XSS : float
        YSS : float
        NODATA : str
        kwargs
        """

        super(AntGainPhaseType, self).__init__(
            Identifier=Identifier, ElementFormat=ElementFormat, X0=X0, Y0=Y0,
            XSS=XSS, YSS=YSS, NODATA=NODATA, **kwargs)


class DwellTimeArrayType(SupportArrayCore):
    """
    Array of COD times (sec) and Dwell Times (sec) for points on reference surface.
    Array coordinates are image area coordinates (IAX, IAY)
    """

    _fields = ('Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS', 'NODATA')
    _required = ('Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS')
    # descriptors
    ElementFormat = StringEnumDescriptor(
        'ElementFormat', ('COD=F4;DT=F4;', ), _required, strict=DEFAULT_STRICT,
        default_value='COD=F4;DT=F4;',
        docstring='The data element format.')  # type: str

    def __init__(
            self,
            Identifier: str = None,
            ElementFormat: str = 'COD=F4;DT=F4;',
            X0: float = None,
            Y0: float = None,
            XSS: float = None,
            YSS: float = None,
            NODATA: str = None,
            **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        X0 : float
        Y0 : float
        XSS : float
        YSS : float
        NODATA : str
        kwargs
        """

        super(DwellTimeArrayType, self).__init__(
            Identifier=Identifier, ElementFormat=ElementFormat, X0=X0, Y0=Y0,
            XSS=XSS, YSS=YSS, NODATA=NODATA, **kwargs)


class AddedSupportArrayType(SupportArrayCore):
    """
    Additional arrays (two-dimensional), where the content and format and units
    of each element are user defined.
    """

    _fields = (
        'Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS', 'NODATA',
        'XUnits', 'YUnits', 'ZUnits', 'Parameters')
    _required = (
        'Identifier', 'ElementFormat', 'X0', 'Y0', 'XSS', 'YSS',
        'XUnits', 'YUnits', 'ZUnits')
    _collections_tags = {
        'Parameters': {'array': False, 'child_tag': 'Parameter'}}
    # descriptors
    XUnits = StringDescriptor(
        'XUnits', _required, strict=DEFAULT_STRICT,
        docstring='The X units.')  # type: str
    YUnits = StringDescriptor(
        'YUnits', _required, strict=DEFAULT_STRICT,
        docstring='The Y units.')  # type: str
    ZUnits = StringDescriptor(
        'ZUnits', _required, strict=DEFAULT_STRICT,
        docstring='The Z units.')  # type: str
    Parameters = ParametersDescriptor(
        'Parameters', _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Other necessary free-form parameters.')  # type: Union[None, ParametersCollection]

    def __init__(self, Identifier=None, ElementFormat=None,
                 X0=None, Y0=None, XSS=None, YSS=None, NODATA=None,
                 XUnits=None, YUnits=None, ZUnits=None, Parameters=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        ElementFormat : str
        X0 : float
        Y0 : float
        XSS : float
        YSS : float
        NODATA : str
        XUnits : str
        YUnits : str
        ZUnits : str
        Parameters : ParametersCollection|dict
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.XUnits = XUnits
        self.YUnits = YUnits
        self.ZUnits = ZUnits
        self.Parameters = Parameters
        super(AddedSupportArrayType, self).__init__(
            Identifier=Identifier, ElementFormat=ElementFormat, X0=X0, Y0=Y0,
            XSS=XSS, YSS=YSS, NODATA=NODATA, **kwargs)


class SupportArrayType(Serializable):
    """
    Parameters that describe the binary support array(s) content and
    grid coordinates.
    """

    _fields = ('IAZArray', 'AntGainPhase', 'DwellTimeArray', 'AddedSupportArray')
    _required = ()
    _collections_tags = {
        'IAZArray': {'array': False, 'child_tag': 'IAZArray'},
        'AntGainPhase': {'array': False, 'child_tag': 'AntGainPhase'},
        'DwellTimeArray': {'array': False, 'child_tag': 'DwellTimeArray'},
        'AddedSupportArray': {'array': False, 'child_tag': 'AddedSupportArray'}}
    # descriptors
    IAZArray = SerializableListDescriptor(
        'IAZArray', IAZArrayType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Arrays of scene surface heights expressed in image coordinate IAZ '
                  'values (meters). Grid coordinates are image area coordinates '
                  '(IAX, IAY).')  # type: Union[None, List[IAZArrayType]]
    AntGainPhase = SerializableListDescriptor(
        'AntGainPhase', AntGainPhaseType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Antenna arrays with values are antenna gain and phase expressed in dB '
                  'and cycles. Array coordinates are direction cosines with respect to '
                  'the ACF (DCX, DCY).')  # type: Union[None, List[AntGainPhaseType]]
    DwellTimeArray = SerializableListDescriptor(
        'DwellTimeArray', DwellTimeArrayType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Array of COD times (sec) and Dwell Times (sec) for points on '
                  'reference surface. Array coordinates are image area '
                  'coordinates (IAX, IAY).')  # type: Union[None, List[DwellTimeArrayType]]
    AddedSupportArray = SerializableListDescriptor(
        'AddedSupportArray', AddedSupportArrayType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Additional arrays (two-dimensional), where the content and format and units of each '
                  'element are user defined.')  # type: Union[None, List[AddedSupportArrayType]]

    def __init__(
            self,
            IAZArray: Optional[List[IAZArrayType]] = None,
            AntGainPhase: Optional[List[AntGainPhaseType]] = None,
            DwellTimeArray: Optional[List[DwellTimeArrayType]] = None,
            AddedSupportArray: Optional[List[AddedSupportArrayType]] = None,
            **kwargs):
        """

        Parameters
        ----------
        IAZArray : None|List[IAZArrayType]
        AntGainPhase : None|List[AntGainPhaseType]
        DwellTimeArray : None|List[DwellTimeArrayType]
        AddedSupportArray : None|List[AddedSupportArrayType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.IAZArray = IAZArray
        self.AntGainPhase = AntGainPhase
        self.DwellTimeArray = DwellTimeArray
        self.AddedSupportArray = AddedSupportArray
        super(SupportArrayType, self).__init__(**kwargs)

    def find_support_array(
            self,
            identifier: str) -> Union[IAZArrayType, AntGainPhaseType, DwellTimeArrayType, AddedSupportArrayType]:
        """
        Find and return the details for support array associated with the given identifier.

        Parameters
        ----------
        identifier : str

        Returns
        -------
        IAZArrayType|AntGainPhaseType|DwellTimeArrayType|AddedSupportArrayType
        """

        if self.IAZArray is not None:
            for entry in self.IAZArray:
                if entry.Identifier == identifier:
                    return entry

        if self.AntGainPhase is not None:
            for entry in self.AntGainPhase:
                if entry.Identifier == identifier:
                    return entry

        if self.DwellTimeArray is not None:
            for entry in self.DwellTimeArray:
                if entry.Identifier == identifier:
                    return entry

        if self.AddedSupportArray is not None:
            for entry in self.AddedSupportArray:
                if entry.Identifier == identifier:
                    return entry

        raise KeyError('Identifier {} not associated with a support array.'.format(identifier))

    def version_required(self) -> Tuple[int, int, int]:
        required = (1, 0, 1)
        if self.DwellTimeArray is not None:
            required = max(required, (1, 1, 0))
        return required



"""
Base common features for product readers
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


from typing import Sequence, List, Tuple, Union, Optional

from sarpy.io.general.base import BaseReader
from sarpy.io.general.data_segment import DataSegment

from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1
from sarpy.io.product.sidd2_elements.SIDD import SIDDType as SIDDType2
from sarpy.io.product.sidd3_elements.SIDD import SIDDType as SIDDType3
from sarpy.io.complex.sicd_elements.SICD import SICDType


class SIDDTypeReader(BaseReader):
    def __init__(self,
                 data_segment: Union[None, DataSegment, Sequence[DataSegment]],
                 sidd_meta: Union[None, SIDDType3, SIDDType2, SIDDType1, Sequence[SIDDType1], Sequence[SIDDType2], Sequence[SIDDType3]],
                 sicd_meta: Union[None, SICDType, Sequence[SICDType]],
                 close_segments: bool = True,
                 delete_files: Union[None, str, Sequence[str]] = None):
        """

        Parameters
        ----------
        data_segment : None|DataSegment|Sequence[DataSegment]
        sidd_meta : None|SIDDType3|SIDDType1|SIDDType2|Sequence[SIDDType3]|Sequence[SIDDType1]|Sequence[SIDDType2]
            The SIDD metadata object(s).
        sicd_meta : None|SICDType|Sequence[SICDType]
            The SICD metadata object(s).
        close_segments : bool
            Call segment.close() for each data segment on reader.close()?
        delete_files : None|Sequence[str]
            Any temp files which should be cleaned up on reader.close()?
            This will occur after closing segments.
        """

        if sidd_meta is None:
            self._sidd_meta = None
        elif isinstance(sidd_meta, (SIDDType1, SIDDType2, SIDDType3)):
            self._sidd_meta = sidd_meta
        else:
            temp_list = []  # type: List[Union[SIDDType1]]
            for el in sidd_meta:
                if not isinstance(el, (SIDDType1, SIDDType2, SIDDType3)):
                    raise TypeError(
                        'Got a collection for sidd_meta, and all elements are required '
                        'to be instances of SIDDType.')
                temp_list.append(el)
            self._sidd_meta = tuple(temp_list)

        if sicd_meta is None:
            self._sicd_meta = None
        elif isinstance(sicd_meta, SICDType):
            self._sicd_meta = (sicd_meta, )
        else:
            temp_list = []  # type: List[SICDType]
            for el in sicd_meta:
                if not isinstance(el, SICDType):
                    raise TypeError(
                        'Got a collection for sicd_meta, and all elements are required '
                        'to be instances of SICDType.')
                temp_list.append(el)
            self._sicd_meta = tuple(temp_list)

        BaseReader.__init__(
            self, data_segment, reader_type='SIDD', close_segments=close_segments, delete_files=delete_files)

    def _check_sizes(self) -> None:
        data_sizes = self.get_data_size_as_tuple()
        sidds = self.get_sidds_as_tuple()
        if len(data_sizes) != len(sidds):
            raise ValueError(
                'Got mismatched number of data segments ({}) and sidds ({})'.format(
                    len(data_sizes), len(sidds)))

        agree = True
        msg = ''
        for i, (data_size, sidd) in enumerate(zip(data_sizes, sidds)):
            expected_size = (sidd.Measurement.PixelFootprint.Row, sidd.Measurement.PixelFootprint.Col)
            if data_size[:2] != expected_size:
                agree = False
                msg += 'data segment at index {} has data size {}\n\t' \
                       'and expected size (from the sidd) {}\n'.format(i, data_size, expected_size)
        if not agree:
            raise ValueError(msg)

    @property
    def sidd_meta(self) -> Union[None, SIDDType1, SIDDType2, SIDDType3, Tuple[SIDDType1, ...], Tuple[SIDDType2, ...], Tuple[SIDDType3, ...]]:
        """
        None|SIDDType1|SIDDType2|SIDDType3|Tuple[SIDDType1, ...]|Tuple[SIDDType2, ...]|Tuple[SIDDType3, ...]: the sidd meta_data collection.
        """

        return self._sidd_meta

    @property
    def sicd_meta(self) -> Optional[Tuple[SICDType, ...]]:
        """
        None|Tuple[SICDType, ...]: the sicd meta_data collection.
        """

        return self._sicd_meta

    def get_sidds_as_tuple(self) -> Union[None, Tuple[SIDDType1, ...], Tuple[SIDDType2, ...], Tuple[SIDDType3, ...]]:
        """
        Get the sidd collection as a tuple - for simplicity and consistency of use.

        Returns
        -------
        None|Tuple[SIDDType1, ...]|Tuple[SIDDType2, ...]|Tuple[SIDDType3, ...]
        """

        if self.sidd_meta is None:
            return None
        elif isinstance(self._sidd_meta, tuple):
            return self.sidd_meta
        else:
            # noinspection PyRedundantParentheses
            return (self.sidd_meta, )



"""
The DataType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import List

from .base import DEFAULT_STRICT
from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import StringDescriptor, StringEnumDescriptor, \
    IntegerDescriptor, SerializableListDescriptor

from .utils import binary_format_string_to_dtype


class ChannelSizeType(Serializable):
    """
    Parameters that define the Channel signal array and PVP array size and location.
    """

    _fields = (
        'Identifier', 'NumVectors', 'NumSamples', 'SignalArrayByteOffset', 'PVPArrayByteOffset',
        'CompressedSignalSize')
    _required = (
        'Identifier', 'NumVectors', 'NumSamples', 'SignalArrayByteOffset', 'PVPArrayByteOffset')
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='String that uniquely identifies the CPHD channel for which the data '
                  'applies.')  # type: str
    NumVectors = IntegerDescriptor(
        'NumVectors', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of vectors in the signal array.')  # type: int
    NumSamples = IntegerDescriptor(
        'NumSamples', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of samples per vector in the signal array.')  # type: int
    SignalArrayByteOffset = IntegerDescriptor(
        'SignalArrayByteOffset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Signal Array offset from the start of the Signal block (in bytes) to the '
                  'start of the Signal Array for the channel.')  # type: int
    PVPArrayByteOffset = IntegerDescriptor(
        'PVPArrayByteOffset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='PVP Array offset from the start of the PVP block (in bytes) to the '
                  'start of the PVP Array for the channel.')  # type: int
    CompressedSignalSize = IntegerDescriptor(
        'CompressedSignalSize', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Size (in bytes) of the compressed signal array byte sequence for the data channel. '
                  'Parameter included if and only if the signal arrays are stored in '
                  'compressed format.')  # type: int

    def __init__(self, Identifier=None, NumVectors=None, NumSamples=None, SignalArrayByteOffset=None,
                 PVPArrayByteOffset=None, CompressedSignalSize=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        NumVectors : int
        NumSamples : int
        SignalArrayByteOffset : int
        PVPArrayByteOffset : int
        CompressedSignalSize : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.NumVectors = NumVectors
        self.NumSamples = NumSamples
        self.SignalArrayByteOffset = SignalArrayByteOffset
        self.PVPArrayByteOffset = PVPArrayByteOffset
        self.CompressedSignalSize = CompressedSignalSize
        super(ChannelSizeType, self).__init__(**kwargs)


class SupportArraySizeType(Serializable):
    """
    Support Array size parameters.
    """

    _fields = ('Identifier', 'NumRows', 'NumCols', 'BytesPerElement', 'ArrayByteOffset')
    _required = _fields
    # descriptors
    Identifier = StringDescriptor(
        'Identifier', _required, strict=DEFAULT_STRICT,
        docstring='Unique string that identifies this support array.')  # type: str
    NumRows = IntegerDescriptor(
        'NumRows', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of rows in the array.')  # type: int
    NumCols = IntegerDescriptor(
        'NumCols', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Number of columns per row in the array.')  # type: int
    BytesPerElement = IntegerDescriptor(
        'BytesPerElement', _required, strict=DEFAULT_STRICT, bounds=(1, None),
        docstring='Indicates the size in bytes of each data element in the support '
                  'array. Each element contains 1 or more binary-formatted '
                  'components.')  # type: int
    ArrayByteOffset = IntegerDescriptor(
        'ArrayByteOffset', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Array offset from the start of the Support block (in bytes) to '
                  'the start of the support array.')  # type: int

    def __init__(self, Identifier=None, NumRows=None, NumCols=None, BytesPerElement=None,
                 ArrayByteOffset=None, **kwargs):
        """

        Parameters
        ----------
        Identifier : str
        NumRows : int
        NumCols : int
        BytesPerElement : int
        ArrayByteOffset : int
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Identifier = Identifier
        self.NumRows = NumRows
        self.NumCols = NumCols
        self.BytesPerElement = BytesPerElement
        self.ArrayByteOffset = ArrayByteOffset
        super(SupportArraySizeType, self).__init__(**kwargs)

    def calculate_size(self):
        """
        Calculates the size of the support array in bytes as described by the contained fields.
        """
        return self.BytesPerElement * self.NumRows * self.NumCols


class DataType(Serializable):
    """
    Parameters that describe binary data components contained in the product.
    """

    _fields = (
        'SignalArrayFormat', 'NumBytesPVP', 'NumCPHDChannels',
        'SignalCompressionID', 'Channels', 'NumSupportArrays', 'SupportArrays')
    _required = ('SignalArrayFormat', 'NumBytesPVP', 'NumCPHDChannels', 'Channels', 'NumSupportArrays')
    _collections_tags = {
        'Channels': {'array': False, 'child_tag': 'Channel'},
        'SupportArrays': {'array': False, 'child_tag': 'SupportArray'}}
    # descriptors
    SignalArrayFormat = StringEnumDescriptor(
        'SignalArrayFormat', ('CI2', 'CI4', 'CF8'), _required, strict=DEFAULT_STRICT,
        docstring='Signal Array sample binary format of the CPHD signal arrays in standard '
                  '(i.e. uncompressed) format, where `CI2` denotes a 1 byte signed integer '
                  "parameter, 2's complement format, and 2 Bytes Per Sample; `CI4` denotes "
                  "a 2 byte signed integer parameter, 2's complement format, and "
                  "4 Bytes Per Sample; `CF8` denotes a 4 byte floating point parameter, and "
                  "8 Bytes Per Sample.")  # type: str
    NumBytesPVP = IntegerDescriptor(
        'NumBytesPVP', _required, strict=DEFAULT_STRICT, bounds=(0, None),
        docstring='Number of bytes per set of Per Vector Parameters, where there is '
                  'one set of PVPs for each CPHD signal vector')  # type: int
    SignalCompressionID = StringDescriptor(
        'SignalCompressionID', _required, strict=DEFAULT_STRICT,
        docstring='Parameter that indicates the signal arrays are in compressed format. Value '
                  'identifies the method of decompression. Parameter included if and only if '
                  'the signal arrays are in compressed format.')  # type: str
    Channels = SerializableListDescriptor(
        'Channels', ChannelSizeType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Parameters that define the Channel signal array and PVP array size '
                  'and location.')  # type: List[ChannelSizeType]
    SupportArrays = SerializableListDescriptor(
        'SupportArrays', SupportArraySizeType, _collections_tags, _required, strict=DEFAULT_STRICT,
        docstring='Support Array size parameters. Branch repeated for each binary support array. '
                  'Support Array referenced by its unique Support Array '
                  'identifier.')  # type: List[SupportArraySizeType]

    def __init__(self, SignalArrayFormat=None, NumBytesPVP=None,
                 SignalCompressionID=None, Channels=None, SupportArrays=None, **kwargs):
        """

        Parameters
        ----------
        SignalArrayFormat : str
        NumBytesPVP : int
        SignalCompressionID : None|str
        Channels : List[ChannelSizeType]
        SupportArrays : None|List[SupportArraySizeType]
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.SignalArrayFormat = SignalArrayFormat
        self.NumBytesPVP = NumBytesPVP
        self.SignalCompressionID = SignalCompressionID
        self.Channels = Channels
        self.SupportArrays = SupportArrays
        super(DataType, self).__init__(**kwargs)

    @property
    def NumSupportArrays(self):
        """
        int: The number of support arrays.
        """

        if self.SupportArrays is None:
            return 0
        else:
            return len(self.SupportArrays)

    @property
    def NumCPHDChannels(self):
        """
        int: The number of CPHD channels.
        """

        if self.Channels is None:
            return 0
        else:
            return len(self.Channels)

    def calculate_support_block_size(self):
        """
        Calculates the size of the support block in bytes as described by the SupportArray fields.
        """
        if self.SupportArrays is None:
            return 0
        else:
            return sum([s.calculate_size() for s in self.SupportArrays])

    def calculate_pvp_block_size(self):
        """
        Calculates the size of the PVP block in bytes as described by the Data fields.
        """
        if self.Channels is None:
            return 0
        else:
            return self.NumBytesPVP * sum([c.NumVectors for c in self.Channels])

    def calculate_signal_block_size(self):
        """
        Calculates the size of the signal block in bytes as described by the Data fields.
        """
        if self.Channels is None:
            return 0

        if self.SignalCompressionID is not None:
            return sum([c.CompressedSignalSize for c in self.Channels])
        else:
            num_bytes_per_sample = binary_format_string_to_dtype(self.SignalArrayFormat).itemsize
            return num_bytes_per_sample * sum([c.NumVectors * c.NumSamples for c in self.Channels])



"""
Remove the existing spectral taper window from a sicd-type, if necessary,
then apply a new spectral taper window.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Valkyrie Systems Corporation"

import copy

import numpy as np
import scipy.fft
import scipy.interpolate as spi

from sarpy.io.complex.sicd_elements.Grid import WgtTypeType
from sarpy.processing.sicd.normalize_sicd import apply_skew_poly
import sarpy.processing.sicd.windows as windows


class Taper:
    """
    This is a helper class that wraps various window function in a common interface.

    Args
    ----
    window: str
        The name of the taper (window) function name.  Acceptable (case insensitive) names are:

        * "uniform": This is a flat spectral taper (i.e., no spectral taper).
        * "hamming": Hamming window taper (0.54 + 0.46 * cos(2*pi*n/M))
        * "hanning" | "hann":  Hann window taper (0.5 + 0.5 * cos(2*pi*n/M))
        * "general_hamming": Raised cosine window (alpha + (1 - alpha) * cos(2*pi*n/M)), default (alpha = 0.5)
        * "kaiser": Kaiser-Bessel window (I0[beta * sqrt(1 - (2*n/M - 1)**2)] / I0[beta]) default (beta = 4)
        * "taylor": Taylor window: Default is a nbar = 4, sll = -30.

    pars: dict | None
        Optional dictionary of parameter values for those windows that have parameter values.
        If omitted or None, the default parameters are used.
        The parameter values for each window are listed above.

        * "uniform": no parameters
        * "hamming": no parameters
        * "hanning" | "hann": no parameters
        * "general_hamming": parameter "alpha"
                 Raised cosine window, w[n] = (alpha + (1 - alpha) * cos(2*pi*n/M)).
                 The default value is alpha = 0.5
        * "kaiser": parameter "beta"
                Kaiser-Bessel window, w[n] = I0[beta * sqrt(1 - (2*n/M - 1)**2)] / I0[beta]
                The default value is beta = 4.
        * "taylor": parameters "nbar" and "sll"
                "nbar" = number of near side lobes and "sll" = side lobe level (dB) of near side lobes.
                The default values are nbar = 4, sll = -30.

    default_size: int
        Optional taper window size.  This argument is almost never needed (see note below).  (default: 513)

    Note
    ----
    This class always creates a symmetric window regardless of whether the window length is odd or even.
    It is not intended that the internal representation of the window samples be used directly.  Instead,
    the "get_vals(size, sym)" method should be used.  Accessing the window values this way will give you
    full control over the size of the window function and whether the window is symmetric.

    """
    def __init__(self, window='UNIFORM', pars=None, default_size=513):
        pars = {key.upper(): val for key, val in pars.items()} if pars else {}    # Force upper-case parameter names

        self.default_pars = {"UNIFORM": {},
                             "HAMMING": {},
                             "HANNING": {},
                             "HANN": {},
                             "GENERAL_HAMMING": {'ALPHA': 0.50},
                             "KAISER": {'BETA': 4.0},
                             "TAYLOR": {"NBAR": 4, "SLL": -30.0},
                             }
        self.default_size = default_size
        self.window_type = window.upper()
        self.window_pars = {**self.default_pars.get(self.window_type, {}), **pars}
        self.window_vals = self._make_sym_1d_window(self.default_size)

    def _make_sym_1d_window(self, window_size=None):
        """Make the sample values of a 1-D symmetric window"""
        window_size = self.default_size if window_size is None else window_size

        if self.window_type == 'UNIFORM':
            wgts = np.ones(window_size)

        elif self.window_type in ['HAMMING']:
            wgts = windows.hamming(window_size, sym=True)

        elif self.window_type in ['HANNING', 'HANN']:
            wgts = windows.hanning(window_size, sym=True)

        elif self.window_type in ['GENERAL_HAMMING']:
            alpha = float(self.window_pars['ALPHA'])
            wgts = windows.general_hamming(window_size, alpha=alpha, sym=True)

        elif self.window_type == 'TAYLOR':
            nbar = int(self.window_pars['NBAR'])
            sll = float(self.window_pars['SLL'])
            wgts = windows.taylor(window_size, nbar=nbar, sll=-np.abs(sll), sym=True)

        elif self.window_type == 'KAISER':
            beta = float(self.window_pars['BETA'])
            wgts = windows.kaiser(window_size, beta=beta, sym=True)

        else:
            raise ValueError(f'Window type "{self.window_type}" is not supported.')

        return wgts

    def get_vals(self, size, sym=True):
        """
        Get interpolated samples of the prototype window.

        Args
        ----
        size: int
            The number of output samples
        sym: bool
            Whether interpolated samples should be symmetric (Default: True)

        Returns
        -------
        vals: numpy.ndarray
            1-D array of sampled taper values.
        """
        return _fit_1d_window(self.window_vals, size) if sym else _fit_1d_window(self.window_vals, size+1)[:-1]


def apply_spectral_taper(sicd_reader, taper):
    """
    Apply a spectral taper window function to SICD image data.
    If necessary, the existing spectral taper window will be removed prior to applying
    the new spectral taper window. To remove an existing spectral taper window without
    applying a new taper, specify a new taper=None.

    Args
    ----
    sicd_reader: sarpy.io.complex.sicd.SICDReader
        A SICD reader object containing both image data and metadata.

    taper: Taper | str | None
        A Taper object specifying the spectral domain taper function to be applied.
        Alternative, a string indicating the taper type.  When a string is provided, a Taper
        object will be created with default parameters for the specified taper type.  Use a
        Taper object if you need to modify the taper parameters.  For example::

            apply_spectral_taper(sicd, Taper(window_name, pars=pars_dict))

    Returns
    -------
    cdata: numpy.ndarray
        The sicd image data modified by removing the existing spectral taper window and applying the specified window.

    sicd_mdata: SICDType
        The modified sicd metadata with updated Grid parameters

    """
    taper = "UNIFORM" if taper is None else taper

    if isinstance(taper, str):
        taper = Taper(taper)

    cdata = sicd_reader[:, :]
    mdata = copy.deepcopy(sicd_reader.sicd_meta)

    cdata = _apply_2d_spectral_taper(cdata, mdata, taper.window_vals)

    # Update the SICD metadata to account for the spectral weighting changes
    row_inv_osf = mdata.Grid.Row.ImpRespBW * mdata.Grid.Row.SS
    col_inv_osf = mdata.Grid.Col.ImpRespBW * mdata.Grid.Col.SS

    old_row_wgts = _get_sicd_wgt_funct(mdata, 'Row', len(taper.window_vals))
    old_col_wgts = _get_sicd_wgt_funct(mdata, 'Col', len(taper.window_vals))
    old_coh_amp_gain = np.mean(old_row_wgts) * np.mean(old_col_wgts) * row_inv_osf * col_inv_osf
    old_rms_amp_gain = np.sqrt(np.mean(old_row_wgts**2) * np.mean(old_col_wgts**2) * row_inv_osf * col_inv_osf)

    new_row_wgts = taper.window_vals
    new_col_wgts = taper.window_vals
    new_coh_amp_gain = np.mean(new_row_wgts) * np.mean(new_col_wgts) * row_inv_osf * col_inv_osf
    new_rms_amp_gain = np.sqrt(np.mean(new_row_wgts**2) * np.mean(new_col_wgts**2) * row_inv_osf * col_inv_osf)

    coh_pwr_gain = (new_coh_amp_gain / old_coh_amp_gain) ** 2
    rms_pwr_gain = (new_rms_amp_gain / old_rms_amp_gain) ** 2

    mdata.Grid.Row.WgtType = WgtTypeType(WindowName=taper.window_type.upper(), Parameters=taper.window_pars)
    mdata.Grid.Col.WgtType = WgtTypeType(WindowName=taper.window_type.upper(), Parameters=taper.window_pars)

    taper_is_uniform = np.all(taper.window_vals == taper.window_vals[0])
    mdata.Grid.Row.WgtFunct = None if taper_is_uniform else taper.window_vals
    mdata.Grid.Col.WgtFunct = None if taper_is_uniform else taper.window_vals

    ipr_half_power_width = windows.find_half_power(taper.window_vals, oversample=16)
    mdata.Grid.Row.ImpRespWid = ipr_half_power_width / mdata.Grid.Row.ImpRespBW
    mdata.Grid.Col.ImpRespWid = ipr_half_power_width / mdata.Grid.Col.ImpRespBW

    if mdata.Radiometric:
        if mdata.Radiometric.NoiseLevel:
            if mdata.Radiometric.NoiseLevel.NoiseLevelType == "ABSOLUTE":
                mdata.Radiometric.NoiseLevel.NoisePoly.Coefs *= rms_pwr_gain

        if mdata.Radiometric.RCSSFPoly:
            mdata.Radiometric.RCSSFPoly.Coefs /= coh_pwr_gain

        if mdata.Radiometric.SigmaZeroSFPoly:
            mdata.Radiometric.SigmaZeroSFPoly.Coefs /= rms_pwr_gain

        if mdata.Radiometric.BetaZeroSFPoly:
            mdata.Radiometric.BetaZeroSFPoly.Coefs /= rms_pwr_gain

        if mdata.Radiometric.GammaZeroSFPoly:
            mdata.Radiometric.GammaZeroSFPoly.Coefs /= rms_pwr_gain

    return cdata, mdata


def _apply_2d_spectral_taper(cdata, mdata, window_vals):
    for axis in ['Row', 'Col']:
        existing_window = _get_sicd_wgt_funct(mdata, axis, len(window_vals))
        both_windows = window_vals / np.maximum(existing_window, 0.01 * np.max(existing_window))

        cdata = _apply_1d_spectral_taper(cdata, mdata, axis, both_windows)

    return cdata


def _apply_1d_spectral_taper(cdata, mdata, axis, window_vals):
    axis_index = {'Row': 0, 'Col': 1}[axis]
    axis_mdata = {'Row': mdata.Grid.Row, 'Col': mdata.Grid.Col}[axis]

    xrow = (np.arange(0, mdata.ImageData.NumRows) - mdata.ImageData.SCPPixel.Row) * mdata.Grid.Row.SS
    ycol = (np.arange(0, mdata.ImageData.NumCols) - mdata.ImageData.SCPPixel.Col) * mdata.Grid.Col.SS

    delta_k_coa_poly = np.array([[0.0]]) if axis_mdata.DeltaKCOAPoly is None else axis_mdata.DeltaKCOAPoly.Coefs

    if not np.all(delta_k_coa_poly == 0):
        cdata = apply_skew_poly(cdata, delta_k_coa_poly, row_array=xrow, col_array=ycol,
                                fft_sgn=axis_mdata.Sgn, dimension=axis_index, forward=False)

    cdata = _fft_window_ifft(cdata, mdata, axis, window_vals)

    if not np.all(delta_k_coa_poly == 0):
        cdata = apply_skew_poly(cdata, delta_k_coa_poly, row_array=xrow, col_array=ycol,
                                fft_sgn=axis_mdata.Sgn, dimension=axis_index, forward=True)
    return cdata


def _get_sicd_wgt_funct(mdata, axis, desired_size=513):
    axis_mdata = {'Row': mdata.Grid.Row, 'Col': mdata.Grid.Col}[axis]

    if axis_mdata.WgtFunct is not None:
        # Get the SICD WgtFunct values and interpolate as needed to produce a window of the desired size.
        wgts = _fit_1d_window(axis_mdata.WgtFunct, desired_size=desired_size)

    else:
        # The SICD WgtFunct values do not exist, so if WindowName is not specified,
        # or the WindowName is "UNIFORM", then return a window of all ones.
        # If WindowName is specified as other than "UNIFORM" then we can not presume to
        # know what the WindowName means because it is not clearly defined in the SICD spec.
        window_name = axis_mdata.WgtType.WindowName if axis_mdata.WgtType else "UNIFORM"

        if window_name.upper() == 'UNIFORM':
            wgts = np.ones(desired_size)
        else:
            raise ValueError(f'SICD/Grid/{axis}/WgtFunct is not part of the SICD metadata, but there appears '
                             f'to be a window of type "{window_name}" applied to the {axis} axis spectrum.')

    return wgts


def _fit_1d_window(window_vals, desired_size):
    if len(window_vals) == desired_size:
        w = window_vals
    else:
        x = np.linspace(0, 1, len(window_vals))
        f = spi.interp1d(x, window_vals, kind='cubic')
        w = f(np.linspace(0, 1, desired_size))
    return w


def _fft_window_ifft(cdata, mdata, axis, window_vals):
    axis_index = {'Row': 0, 'Col': 1}[axis]
    axis_mdata = {'Row': mdata.Grid.Row, 'Col': mdata.Grid.Col}[axis]
    nyquist_bw = 1.0 / axis_mdata.SS
    ipr_bw = axis_mdata.ImpRespBW
    osf = nyquist_bw / ipr_bw

    # Zero pad the image data to avoid IPR wrap-around, then
    # find a good FFT size which creates some additional zero pad.
    axis_size = cdata.shape[axis_index]
    wrap_around_pad = int(min(200 * osf, 0.1 * axis_size))
    good_fft_size = scipy.fft.next_fast_len(axis_size + wrap_around_pad)

    # Forward transform without FFTSHIFT so the DC bin is at index=0
    if axis_mdata.Sgn == -1:
        cdata_fft = scipy.fft.fft(cdata, n=good_fft_size, axis=axis_index)
    else:
        cdata_fft = scipy.fft.ifft(cdata, n=good_fft_size, axis=axis_index)

    # Interpolate the taper to cover the spectral support bandwidth and extend the
    # taper window's end points into the over sample region of the spectrum.
    f = spi.interp1d(np.linspace(-ipr_bw / 2, ipr_bw / 2, len(window_vals)), window_vals, kind='cubic',
                     bounds_error=False, fill_value=(window_vals[0], window_vals[-1]))
    padded_taper = f(scipy.fft.fftfreq(good_fft_size, axis_mdata.SS))

    # Apply the taper to the spectrum.
    taper_2d = padded_taper[:, np.newaxis] if axis == 'Row' else padded_taper[np.newaxis, :]
    cdata_fft *= taper_2d

    # Inverse transform without FFTSHIFT and trim back to the original image size.
    nrows, ncols = cdata.shape
    if axis_mdata.Sgn == -1:
        cdata = scipy.fft.ifft(cdata_fft, n=good_fft_size, axis=axis_index)[:nrows, :ncols]
    else:
        cdata = scipy.fft.fft(cdata_fft, n=good_fft_size, axis=axis_index)[:nrows, :ncols]

    return cdata



"""
Common use sicd_elements methods.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
from typing import Optional, Tuple
import numpy
from sarpy.io.general.utils import get_seconds


logger = logging.getLogger(__name__)


def _get_center_frequency(
        RadarCollection,
        ImageFormation) -> Optional[float]:
    """
    Helper method.

    Parameters
    ----------
    RadarCollection : sarpy.io.complex.sicd_elements.RadarCollection.RadarCollectionType
    ImageFormation : sarpy.io.complex.sicd_elements.ImageFormation.ImageFormationType

    Returns
    -------
    None|float
        The center processed frequency, in the event that RadarCollection.RefFreqIndex is `None` or `0`.
    """
    if RadarCollection is None or ImageFormation is None or ImageFormation.TxFrequencyProc is None:
        return None
    if RadarCollection.RefFreqIndex is None or RadarCollection.RefFreqIndex == 0:
        return ImageFormation.TxFrequencyProc.center_frequency
    else:
        return None


def polstring_version_required(str_in: Optional[str]) -> Tuple[int, int, int]:
    """
    What SICD version does the pol string require?

    Parameters
    ----------
    str_in : None|str
        The tx/rcv polarization string.

    Returns
    -------
    tuple
        One of `(1, 1, 0)`, `(1, 2, 1)`, `(1, 3, 0)`
    """

    if str_in is None or str_in in ['OTHER', 'UNKNOWN']:
        return (1, 1, 0)

    parts = str_in.split(':')
    if len(parts) != 2:
        logger.error('Expected polarization string of length 2, '
                    'but populated as `{}`'.format(len(parts)))
        return None

    part1, part2 = parts
    if part1.startswith('OTHER') or part2.startswith('OTHER'):
        return (1, 3, 0)
    if part1 in ['S', 'E', 'X', 'Y'] or part2 in ['S', 'E', 'X', 'Y']:
        return (1, 3, 0)
    elif (part1 in ['V', 'H'] and part2 in ['RHC', 'LHC']) or \
            (part2 in ['V', 'H'] and part1 in ['RHC', 'LHC']):
        return (1, 2, 1)
    else:
        return (1, 1, 0)


################
# SICD comparison and matching methods

def is_same_size(sicd1, sicd2) -> bool:
    """
    Are the two SICD structures the same size in pixels?

    Parameters
    ----------
    sicd1 : sarpy.io.complex.sicd_elements.SICD.SICDType
    sicd2 : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if sicd1 is sicd2:
        return True

    try:
        return (sicd1.ImageData.NumRows == sicd2.ImageData.NumRows) and \
               (sicd1.ImageData.NumCols == sicd2.ImageData.NumCols)
    except AttributeError:
        return False


def is_same_sensor(sicd1, sicd2) -> bool:
    """
    Are the two SICD structures from the same sensor?

    Parameters
    ----------
    sicd1 : sarpy.io.complex.sicd_elements.SICD.SICDType
    sicd2 : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if sicd1 is sicd2:
        return True

    try:
        return sicd1.CollectionInfo.CollectorName == sicd2.CollectionInfo.CollectorName
    except AttributeError:
        return False


def is_same_start_time(sicd1, sicd2) -> bool:
    """
    Do the two SICD structures have the same start time with millisecond resolution?

    Parameters
    ----------
    sicd1 : sarpy.io.complex.sicd_elements.SICD.SICDType
    sicd2 : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if sicd1 is sicd2:
        return True

    try:
        return abs(get_seconds(sicd1.Timeline.CollectStart, sicd2.Timeline.CollectStart, precision='ms')) < 2e-3
    except AttributeError:
        return False


def is_same_duration(sicd1, sicd2) -> bool:
    """
    Do the two SICD structures have the same duration, with millisecond resolution?

    Parameters
    ----------
    sicd1 : sarpy.io.complex.sicd_elements.SICD.SICDType
    sicd2 : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if sicd1 is sicd2:
        return True

    try:
        return abs(sicd1.Timeline.CollectDuration - sicd2.Timeline.CollectDuration) < 2e-3
    except AttributeError:
        return False


def is_same_band(sicd1, sicd2) -> bool:
    """
    Are the two SICD structures the same band?

    Parameters
    ----------
    sicd1 : sarpy.io.complex.sicd_elements.SICD.SICDType
    sicd2 : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if sicd1 is sicd2:
        return True

    try:
        return abs(sicd1.Grid.Row.KCtr - sicd2.Grid.Row.KCtr) <= 1./(sicd1.Grid.Row.SS*sicd1.ImageData.NumRows)
    except AttributeError:
        return False


def is_same_scp(sicd1, sicd2) -> bool:
    """
    Do the two SICD structures share the same SCP, with resolution of one meter
    in each ECF coordinate?

    Parameters
    ----------
    sicd1 : sarpy.io.complex.sicd_elements.SICD.SICDType
    sicd2 : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if sicd1 is sicd2:
        return True

    try:
        ecf1 = sicd1.GeoData.SCP.ECF.get_array()
        ecf2 = sicd2.GeoData.SCP.ECF.get_array()
        return numpy.all(numpy.abs(ecf1 - ecf2) < 1)
    except AttributeError:
        return False


def is_general_match(sicd1, sicd2) -> bool:
    """
    Do the two SICD structures seem to form a basic match? This necessarily
    establishes and equivalence relation between sicds.

    Parameters
    ----------
    sicd1 : sarpy.io.complex.sicd_elements.SICD.SICDType
    sicd2 : sarpy.io.complex.sicd_elements.SICD.SICDType

    Returns
    -------
    bool
    """

    if sicd1 is sicd2:
        return True

    return is_same_size(sicd1, sicd2) and is_same_sensor(sicd1, sicd2) and \
        is_same_start_time(sicd1, sicd2) and is_same_duration(sicd1, sicd2) and \
        is_same_band(sicd1, sicd2) and is_same_scp(sicd1, sicd2)



"""
Methods for ortho-rectification
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
from typing import Union

import numpy
from scipy.interpolate import RectBivariateSpline

from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.base import SICDTypeReader

from sarpy.io.complex.sicd_elements.blocks import Poly2DType
from sarpy.geometry.geometry_elements import GeometryObject

from .projection_helper import ProjectionHelper, PGProjection, PGRatPolyProjection
from ..rational_polynomial import SarpyRatPolyError

logger = logging.getLogger(__name__)


def _linear_fill(pixel_array, fill_interval=1):
    """
    This is to fill in linear features in pixel space at the given interval.
    When the first and last vertices are not identical, an implicit edge connecting them is also filled.

    Parameters
    ----------
    pixel_array : numpy.ndarray
    fill_interval : int|float

    Returns
    -------
    numpy.ndarray
    """

    if not isinstance(pixel_array, numpy.ndarray):
        raise TypeError('pixel_array must be a numpy array. Got type {}'.format(type(pixel_array)))
    if pixel_array.ndim < 2:
        # nothing to be done
        return pixel_array

    if pixel_array.ndim > 2:
        raise ValueError('pixel_array must be no more than two-dimensional. Got shape {}'.format(pixel_array.shape))
    if pixel_array.shape[1] != 2:
        raise ValueError(
            'pixel_array is two dimensional, and the final dimension must have length 2. '
            'Got shape {}'.format(pixel_array.shape))
    if pixel_array.shape[0] < 2:
        # nothing to be done
        return pixel_array

    def make_segment(start_point, end_point):
        segment_length = numpy.linalg.norm(end_point - start_point)
        segment_count = max(2, int(numpy.ceil(segment_length/float(fill_interval) + 1)))
        segment = numpy.zeros((segment_count, 2), dtype=numpy.float64)
        # NB: it's ok if start == end in linspace
        segment[:, 0] = numpy.linspace(start_point[0], end_point[0], segment_count)
        segment[:, 1] = numpy.linspace(start_point[1], end_point[1], segment_count)
        return segment

    segments = []
    for i in range(pixel_array.shape[0]-1):
        start_segment = pixel_array[i, :]
        end_segment = pixel_array[i+1, :]
        segments.append(make_segment(start_segment, end_segment))
    if not numpy.array_equal(pixel_array[0], pixel_array[-1]):
        # include line connecting  first and last points if they are not the same
        segments.append(make_segment(pixel_array[0], pixel_array[-1]))
    return numpy.vstack(segments)


class OrthorectificationHelper(object):
    """
    Abstract helper class which defines ortho-rectification process for a sicd-type
    reader object.
    """

    __slots__ = (
        '_reader', '_index', '_sicd', '_proj_helper', '_out_dtype', '_complex_valued',
        '_pad_value', '_apply_radiometric', '_subtract_radiometric_noise',
        '_rad_poly', '_noise_poly', '_default_physical_bounds')

    def __init__(self, reader, index=0, proj_helper=None, complex_valued=False,
                 pad_value=None, apply_radiometric=None, subtract_radiometric_noise=False):
        """

        Parameters
        ----------
        reader : SICDTypeReader
        index : int
        proj_helper : None|ProjectionHelper
            If `None`, this will default to `PGRatPolyProjection(<sicd>)` unless there is
            a SarpyRatPolyError, when it will fall back to `PGProjection(<sicd>)`,
            where `<sicd>` will be the sicd from `reader` at `index`. Otherwise,
            it is the user's responsibility to ensure that `reader`, `index` and
            `proj_helper` are in sync.
        complex_valued : bool
            Do we want complex values returned? If `False`, the magnitude values
            will be used.
        pad_value : None|Any
            Value to use for any out-of-range pixels. Defaults to `0` if not provided.
        apply_radiometric : None|str
            If provided, must be one of `['RCS', 'Sigma0', 'Gamma0', 'Beta0']`
            (not case-sensitive). This will apply the given radiometric scale factor
            to calculated pixel power, with noise subtracted if `subtract_radiometric_noise = True`.
            **Only valid if `complex_valued=False`**.
        subtract_radiometric_noise : bool
            This indicates whether the radiometric noise should be subtracted from
            the pixel amplitude. **Only valid if `complex_valued=False`**.
        """

        self._index = None
        self._sicd = None
        self._proj_helper = None
        self._apply_radiometric = None
        self._subtract_radiometric_noise = None
        self._rad_poly = None  # type: [None, Poly2DType]
        self._noise_poly = None  # type: [None, Poly2DType]
        self._default_physical_bounds = None

        self._pad_value = pad_value
        self._complex_valued = complex_valued
        if self._complex_valued:
            self._out_dtype = numpy.dtype('complex64')
        else:
            self._out_dtype = numpy.dtype('float32')
        if not isinstance(reader, SICDTypeReader):
            raise TypeError('Got unexpected type {} for reader'.format(type(reader)))
        self._reader = reader
        self.apply_radiometric = apply_radiometric
        self.subtract_radiometric_noise = subtract_radiometric_noise
        self.set_index_and_proj_helper(index, proj_helper=proj_helper)

    @property
    def reader(self):
        # type: () -> SICDTypeReader
        """
        SICDTypeReader: The reader instance.
        """

        return self._reader

    @property
    def index(self):
        # type: () -> int
        """
        int: The index for the desired sicd element.
        """
        return self._index

    @property
    def sicd(self):
        # type: () -> SICDType
        """
        SICDType: The sicd structure.
        """

        return self._sicd

    @property
    def proj_helper(self):
        # type: () -> ProjectionHelper
        """
        ProjectionHelper: The projection helper instance.
        """

        return self._proj_helper

    @property
    def out_dtype(self):
        # type: () -> numpy.dtype
        """
        numpy.dtype: The output data type.
        """

        return self._out_dtype

    @property
    def pad_value(self):
        """
        The value to use for any portions of the array which extend beyond the range
        of where the reader has data.
        """

        return self._pad_value

    @pad_value.setter
    def pad_value(self, value):
        self._pad_value = value

    def set_index_and_proj_helper(self, index, proj_helper=None):
        """
        Sets the index and proj_helper objects.

        Parameters
        ----------
        index : int
        proj_helper : ProjectionHelper

        Returns
        -------
        None
        """

        self._index = index
        self._sicd = self.reader.get_sicds_as_tuple()[index]
        self._is_radiometric_valid()
        self._is_radiometric_noise_valid()

        default_ortho_bounds = None
        if proj_helper is None:
            try:
                proj_helper = PGRatPolyProjection(self.sicd)
            except SarpyRatPolyError:
                proj_helper = PGProjection(self.sicd)

            if self.sicd.RadarCollection is not None and self.sicd.RadarCollection.Area is not None \
                    and self.sicd.RadarCollection.Area.Plane is not None:
                plane = self.sicd.RadarCollection.Area.Plane
                default_ortho_bounds = numpy.array([
                    plane.XDir.FirstLine, plane.XDir.FirstLine+plane.XDir.NumLines,
                    plane.YDir.FirstSample, plane.YDir.FirstSample+plane.YDir.NumSamples], dtype=numpy.uint32)

        if not isinstance(proj_helper, ProjectionHelper):
            raise TypeError('Got unexpected type {} for proj_helper'.format(proj_helper))
        self._proj_helper = proj_helper
        if default_ortho_bounds is not None:
            _, ortho_rectangle = self.bounds_to_rectangle(default_ortho_bounds)
            self._default_physical_bounds = self.proj_helper.ortho_to_ecf(ortho_rectangle)

    @property
    def apply_radiometric(self):
        # type: () -> Union[None, str]
        """
        None|str: This indicates which, if any, of the radiometric scale factors
        to apply in the result. If not `None`, this must be one of ('RCS', 'SIGMA0', 'GAMMA0', 'BETA0').

        Setting to a value other than `None` will result in an error if 1.) `complex_valued` is `True`, or
        2.) the appropriate corresponding element `sicd.Radiometric.RCSSFPoly`,
        `sicd.Radiometric.SigmaZeroSFPoly`, `sicd.Radiometric.GammaZeroSFPoly`, or
        `sicd.Radiometric.BetaZeroSFPoly` is not populated with a valid polynomial.
        """

        return self._apply_radiometric

    @apply_radiometric.setter
    def apply_radiometric(self, value):
        if value is None:
            self._apply_radiometric = None
        elif isinstance(value, str):
            val = value.upper()
            allowed = ('RCS', 'SIGMA0', 'GAMMA0', 'BETA0')
            if val not in allowed:
                raise ValueError('Require that value is one of {}, got {}'.format(allowed, val))
            self._apply_radiometric = val
            self._is_radiometric_valid()
        else:
            raise TypeError('Got unexpected type {} for apply_radiometric'.format(type(value)))

    @property
    def subtract_radiometric_noise(self):
        """
        bool: This indicates whether the radiometric noise should be subtracted from
        the pixel amplitude. If `apply_radiometric` is not `None`, then this subtraction
        will happen applying the corresponding scaling.

        Setting this to `True` will **result in an error** unless the given sicd structure has
        `sicd.Radiometric.NoiseLevel.NoisePoly` populated with a viable polynomial and
        `sicd.Radiometric.NoiseLevel.NoiseLevelType == 'ABSOLUTE'`.
        """

        return self._subtract_radiometric_noise

    @subtract_radiometric_noise.setter
    def subtract_radiometric_noise(self, value):
        if value:
            self._subtract_radiometric_noise = True
        else:
            self._subtract_radiometric_noise = False
        self._is_radiometric_noise_valid()

    def _is_radiometric_valid(self):
        """
        Checks whether the apply radiometric settings are valid.

        Returns
        -------
        None
        """

        if self.apply_radiometric is None:
            self._rad_poly = None
            return  # nothing to be done
        if self._complex_valued:
            raise ValueError('apply_radiometric is not None, which requires real valued output.')
        if self.sicd is None:
            return  # nothing to be done, no sicd set (yet)

        if self.sicd.Radiometric is None:
            raise ValueError(
                'apply_radiometric is {}, but sicd.Radiometric is unpopulated.'.format(self.apply_radiometric))

        if self.apply_radiometric == 'RCS':
            if self.sicd.Radiometric.RCSSFPoly is None:
                raise ValueError('apply_radiometric is "RCS", but the sicd.Radiometric.RCSSFPoly is not populated.')
            else:
                self._rad_poly = self.sicd.Radiometric.RCSSFPoly
        elif self.apply_radiometric == 'SIGMA0':
            if self.sicd.Radiometric.SigmaZeroSFPoly is None:
                raise ValueError(
                    'apply_radiometric is "SIGMA0", but the sicd.Radiometric.SigmaZeroSFPoly is not populated.')
            else:
                self._rad_poly = self.sicd.Radiometric.SigmaZeroSFPoly
        elif self.apply_radiometric == 'GAMMA0':
            if self.sicd.Radiometric.GammaZeroSFPoly is None:
                raise ValueError(
                    'apply_radiometric is "GAMMA0", but the sicd.Radiometric.GammaZeroSFPoly is not populated.')
            else:
                self._rad_poly = self.sicd.Radiometric.GammaZeroSFPoly
        elif self.apply_radiometric == 'BETA0':
            if self.sicd.Radiometric.BetaZeroSFPoly is None:
                raise ValueError(
                    'apply_radiometric is "BETA0", but the sicd.Radiometric.BetaZeroSFPoly is not populated.')
            else:
                self._rad_poly = self.sicd.Radiometric.BetaZeroSFPoly
        else:
            raise ValueError('Got unhandled value {} for apply_radiometric'.format(self.apply_radiometric))

    def _is_radiometric_noise_valid(self):
        """
        Checks whether the subtract_radiometric_noise setting is valid.

        Returns
        -------
        None
        """

        if not self.subtract_radiometric_noise:
            self._noise_poly = None
            return  # nothing to be done
        if self._complex_valued:
            raise ValueError('subtract_radiometric_noise is True, which requires real valued output.')
        if self.sicd is None:
            return  # nothing to be done, no sicd set (yet)

        # set the noise polynomial value
        if self.sicd.Radiometric is None:
            raise ValueError(
                'subtract_radiometric_noise is True,\n\t'
                'but sicd.Radiometric is unpopulated.')

        if self.sicd.Radiometric.NoiseLevel is None:
            raise ValueError(
                'subtract_radiometric_noise is set to True,\n\t'
                'but sicd.Radiometric.NoiseLevel is not populated.')
        if self.sicd.Radiometric.NoiseLevel.NoisePoly is None:
            raise ValueError(
                'subtract_radiometric_noise is set to True,\n\t'
                'but sicd.Radiometric.NoiseLevel.NoisePoly is not populated.')
        if self.sicd.Radiometric.NoiseLevel.NoiseLevelType == 'RELATIVE':
            raise ValueError(
                'subtract_radiometric_noise is set to True,\n\t'
                'but sicd.Radiometric.NoiseLevel.NoiseLevelType is "RELATIVE"')
        self._noise_poly = self.sicd.Radiometric.NoiseLevel.NoisePoly

    def get_full_ortho_bounds(self):
        """
        Gets the bounds for the ortho-rectified coordinates for the full sicd image.

        Returns
        -------
        numpy.ndarray
            Of the form `[min row, max row, min column, max column]`.
        """

        if self._default_physical_bounds is not None:
            ortho_rectangle = self.proj_helper.ecf_to_ortho(self._default_physical_bounds)
            return self.proj_helper.get_pixel_array_bounds(ortho_rectangle)

        full_coords = self.sicd.ImageData.get_full_vertex_data()
        full_line = _linear_fill(full_coords, fill_interval=1)
        return self.get_orthorectification_bounds_from_pixel_object(full_line)

    def get_valid_ortho_bounds(self):
        """
        Gets the bounds for the ortho-rectified coordinates for the valid portion
        of the sicd image. This is the outer bounds of the valid portion, so may contain
        some portion which is not itself valid.

        If sicd.ImageData.ValidData is not defined, then the full image bounds will
        be returned.

        Returns
        -------
        numpy.ndarray
            Of the form `[min row, max row, min column, max column]`.
        """

        if self._default_physical_bounds is not None:
            ortho_rectangle = self.proj_helper.ecf_to_ortho(self._default_physical_bounds)
            return self.proj_helper.get_pixel_array_bounds(ortho_rectangle)

        valid_coords = self.sicd.ImageData.get_valid_vertex_data()
        if valid_coords is None:
            valid_coords = self.sicd.ImageData.get_full_vertex_data()
        valid_line = _linear_fill(valid_coords, fill_interval=1)
        return self.get_orthorectification_bounds_from_pixel_object(valid_line)

    def get_orthorectification_bounds_from_pixel_object(self, coordinates):
        """
        Determine the ortho-rectified (coordinate-system aligned) rectangular bounding
        region which contains the provided coordinates in pixel space.

        Parameters
        ----------
        coordinates : GeometryObject|numpy.ndarray|list|tuple
            The coordinate system of the input will be assumed to be pixel space.
            When the first and last vertices are not identical, an implicit edge connecting them is assumed.

        Returns
        -------
        numpy.ndarray
            Of the form `(row_min, row_max, col_min, col_max)`.
        """

        if isinstance(coordinates, GeometryObject):
            # pixel_bounds.shape is (min/max, coordinate-dimensionality)
            pixel_bounds = numpy.array(coordinates.get_bbox()).reshape((2, -1))
            # assume first two coordinate dims are row/col and ignore the rest
            (minrow, mincol), (maxrow, maxcol) = pixel_bounds[:, :2]
            coordinates = numpy.array(
                [[minrow, mincol],
                 [maxrow, mincol],
                 [maxrow, maxcol],
                 [minrow, maxcol]], dtype=numpy.float64)
        filled_coordinates = _linear_fill(coordinates, fill_interval=1)
        ortho = self.proj_helper.pixel_to_ortho(filled_coordinates)
        return self.proj_helper.get_pixel_array_bounds(ortho)

    def get_orthorectification_bounds_from_latlon_object(self, coordinates):
        """
        Determine the ortho-rectified (coordinate-system aligned) rectangular bounding
        region which contains the provided coordinates in lat/lon space.

        Parameters
        ----------
        coordinates : GeometryObject|numpy.ndarray|list|tuple
            The coordinate system of the input will be assumed to be lat/lon space.
            **Note** a GeometryObject is expected to follow lon/lat ordering paradigm,
            by convention.

        Returns
        -------
        numpy.ndarray
            Of the form `(row_min, row_max, col_min, col_max)`.
        """

        if isinstance(coordinates, GeometryObject):
            # Note we assume a geometry object is using lon/lat ordering of coordinates.
            bounds = coordinates.get_bbox()
            if len(bounds) == 4:
                coordinates = numpy.array(
                    [[bounds[1], bounds[0]],
                     [bounds[1], bounds[2]],
                     [bounds[3], bounds[2]],
                     [bounds[3], bounds[0]]], dtype=numpy.float64)
            elif len(bounds) >= 6:
                siz = int(len(bounds)/2)
                coordinates = numpy.array(
                    [[bounds[1], bounds[0], bounds[3]],
                     [bounds[1], bounds[siz], bounds[3]],
                     [bounds[3], bounds[2], bounds[3]],
                     [bounds[3], bounds[0], bounds[3]]], dtype=numpy.float64)
            else:
                raise ValueError(
                    'It is expected that the geometry object "coordinates" uses two '
                    'or three dimensional coordinates. Got {} for a bounding box.'.format(bounds))
        if not isinstance(coordinates, numpy.ndarray):
            coordinates = numpy.array(coordinates, dtype=numpy.float64)
        if coordinates.shape[-1] == 2:
            ortho = self.proj_helper.ll_to_ortho(coordinates)
        elif coordinates.shape[-1] == 3:
            ortho = self.proj_helper.llh_to_ortho(coordinates)
        else:
            raise ValueError('Got unexpected shape for coordinates {}'.format(coordinates.shape))
        return self.proj_helper.get_pixel_array_bounds(ortho)

    @staticmethod
    def validate_bounds(bounds):
        """
        Validate a pixel type bounds array.

        Parameters
        ----------
        bounds : numpy.ndarray|list|tuple

        Returns
        -------
        numpy.ndarray
        """

        if not isinstance(bounds, numpy.ndarray):
            bounds = numpy.asarray(bounds)

        if bounds.ndim != 1 or bounds.size != 4:
            raise ValueError(
                'bounds is required to be one-dimensional and size 4. '
                'Got input shape {}'.format(bounds.shape))
        if bounds[0] >= bounds[1] or bounds[2] >= bounds[3]:
            raise ValueError(
                'bounds is required to be of the form (min row, max row, min col, max col), '
                'got {}'.format(bounds))

        if issubclass(bounds.dtype.type, numpy.integer):
            return bounds
        else:
            out = numpy.zeros((4,), dtype=numpy.int32)
            out[0:3:2] = (numpy.floor(bounds[0]), numpy.floor(bounds[2]))
            out[1:4:2] = (numpy.ceil(bounds[1]), numpy.ceil(bounds[3]))
            return out

    @staticmethod
    def _get_ortho_mesh(ortho_bounds):
        """
        Fetch a the grid of rows/columns coordinates for the desired rectangle.

        Parameters
        ----------
        ortho_bounds : numpy.ndarray
            Of the form `(min row, max row, min col, max col)`.

        Returns
        -------
        numpy.ndarray
        """

        ortho_shape = (int(ortho_bounds[1]-ortho_bounds[0]), int(ortho_bounds[3]-ortho_bounds[2]), 2)
        ortho_mesh = numpy.zeros(ortho_shape, dtype=numpy.int32)
        ortho_mesh[:, :, 1], ortho_mesh[:, :, 0] = numpy.meshgrid(numpy.arange(ortho_bounds[2], ortho_bounds[3]),
                                                                  numpy.arange(ortho_bounds[0], ortho_bounds[1]))
        return ortho_mesh

    @staticmethod
    def _get_mask(pixel_rows, pixel_cols, row_array, col_array):
        """
        Construct the valid mask. This is a helper function, and no error checking
        will be performed for any issues.

        Parameters
        ----------
        pixel_rows : numpy.ndarray
            The array of pixel rows. Must be the same shape as `pixel_cols`.
        pixel_cols : numpy.ndarray
            The array of pixel columns. Must be the same shape as `pixel_rows'.
        row_array : numpy.ndarray
            The rows array used for bounds, must be one dimensional and monotonic.
        col_array : numpy.ndarray
            The columns array used for bounds, must be one dimensional and monotonic.

        Returns
        -------
        numpy.ndarray
        """

        mask = (numpy.isfinite(pixel_rows) &
                numpy.isfinite(pixel_cols) &
                (pixel_rows >= row_array[0]) & (pixel_rows < row_array[-1]) &
                (pixel_cols >= col_array[0]) & (pixel_cols < col_array[-1]))
        return mask

    def bounds_to_rectangle(self, bounds):
        """
        From a bounds style array, construct the four corner coordinate array.
        This follows the SICD convention of going CLOCKWISE around the corners.

        Parameters
        ----------
        bounds : numpy.ndarray|list|tuple
            Of the form `(row min, row max, col min, col max)`.

        Returns
        -------
        (numpy.ndarray, numpy.ndarray)
            The (integer valued) bounds and rectangular coordinates.
        """

        bounds = self.validate_bounds(bounds)
        coords = numpy.zeros((4, 2), dtype=numpy.int32)
        coords[0, :] = (bounds[0], bounds[2])  # row min, col min
        coords[1, :] = (bounds[0], bounds[3])  # row min, col max
        coords[2, :] = (bounds[1], bounds[3])  # row max, col max
        coords[3, :] = (bounds[1], bounds[2])  # row max, col min
        return bounds, coords

    def extract_pixel_bounds(self, bounds):
        """
        Validate the bounds array of orthorectified pixel coordinates, and determine
        the required bounds in reader pixel coordinates. If the

        Parameters
        ----------
        bounds : numpy.ndarray|list|tuple

        Returns
        -------
        (numpy.ndarray, numpy.ndarray)
            The integer valued orthorectified and reader pixel coordinate bounds.
        """

        bounds, coords = self.bounds_to_rectangle(bounds)
        filled_coords = _linear_fill(coords, fill_interval=1)
        pixel_coords = self.proj_helper.ortho_to_pixel(filled_coords)
        pixel_bounds = self.proj_helper.get_pixel_array_bounds(pixel_coords)
        return bounds, self.validate_bounds(pixel_bounds)

    def _initialize_workspace(self, ortho_bounds, final_dimension=0):
        """
        Initialize the orthorectification array workspace.

        Parameters
        ----------
        ortho_bounds : numpy.ndarray
            Of the form `(min row, max row, min col, max col)`.
        final_dimension : int
            The size of the third dimension. If `0`, then it will be omitted.

        Returns
        -------
        numpy.ndarray
        """

        if final_dimension > 0:
            out_shape = (
                int(ortho_bounds[1] - ortho_bounds[0]),
                int(ortho_bounds[3] - ortho_bounds[2]),
                int(final_dimension))
        else:
            out_shape = (
                int(ortho_bounds[1]-ortho_bounds[0]),
                int(ortho_bounds[3]-ortho_bounds[2]))
        return numpy.zeros(out_shape, dtype=self.out_dtype) if self._pad_value is None else \
            numpy.full(out_shape, self._pad_value, dtype=self.out_dtype)

    def get_real_pixel_bounds(self, pixel_bounds):
        """
        Fetch the real pixel limit from the nominal pixel limits - this just
        factors in the image reader extent.

        Parameters
        ----------
        pixel_bounds : numpy.ndarray

        Returns
        -------
        numpy.ndarray
        """

        if pixel_bounds[0] > pixel_bounds[1] or pixel_bounds[2] > pixel_bounds[3]:
            raise ValueError('Got unexpected and invalid pixel_bounds array {}'.format(pixel_bounds))

        pixel_limits = self.reader.get_data_size_as_tuple()[self.index]
        if (pixel_bounds[0] >= pixel_limits[0]) or (pixel_bounds[1] < 0) or \
                (pixel_bounds[2] >= pixel_limits[1]) or (pixel_bounds[3] < 0):
            # this entirely misses the whole region
            return numpy.array([0, 0, 0, 0], dtype=numpy.int32)

        real_pix_bounds = numpy.array([
            max(0, pixel_bounds[0]), min(pixel_limits[0], pixel_bounds[1]),
            max(0, pixel_bounds[2]), min(pixel_limits[1], pixel_bounds[3])], dtype=numpy.int32)
        return real_pix_bounds

    def _apply_radiometric_params(self, pixel_rows, pixel_cols, value_array):
        """
        Apply the radiometric parameters to the solution array.

        Parameters
        ----------
        pixel_rows : numpy.ndarray
        pixel_cols : numpy.ndarray
        value_array : numpy.ndarray

        Returns
        -------
        numpy.ndarray
        """

        if self._rad_poly is None and self._noise_poly is None:
            # nothing to be done.
            return value_array

        if pixel_rows.shape == value_array.shape and pixel_cols.shape == value_array.shape:
            rows_meters = (pixel_rows - self.sicd.ImageData.SCPPixel.Row)*self.sicd.Grid.Row.SS
            cols_meters = (pixel_cols - self.sicd.ImageData.SCPPixel.Col)*self.sicd.Grid.Col.SS
        elif value_array.ndim == 2 and \
            (pixel_rows.ndim == 1 and pixel_rows.size == value_array.shape[0]) and \
                (pixel_cols.ndim == 1 and pixel_cols.size == value_array.shape[1]):
            cols_meters, rows_meters = numpy.meshgrid(
                (pixel_cols - self.sicd.ImageData.SCPPixel.Col)*self.sicd.Grid.Col.SS,
                (pixel_rows - self.sicd.ImageData.SCPPixel.Row) * self.sicd.Grid.Row.SS)
        else:
            raise ValueError(
                'Either pixel_rows, pixel_cols, and value_array must all have the same shape, '
                'or pixel_rows/pixel_cols are one dimensional and '
                'value_array.shape = (pixel_rows.size, pixel_cols.size). Got shapes {}, {}, and '
                '{}'.format(pixel_rows.shape, pixel_cols.shape, value_array.shape))

        # calculate pixel power, with noise subtracted if necessary
        if self._noise_poly is not None:
            noise = numpy.exp(10 * self._noise_poly(rows_meters, cols_meters))  # convert from db to power
            pixel_power = value_array*value_array - noise
            del noise
        else:
            pixel_power = value_array*value_array

        if self._rad_poly is None:
            return numpy.sqrt(pixel_power)
        else:
            return pixel_power*self._rad_poly(rows_meters, cols_meters)

    def _validate_row_col_values(self, row_array, col_array, value_array, value_is_flat=False):
        """
        Helper method for validating the types and shapes.

        Parameters
        ----------
        row_array : numpy.ndarray
            The rows of the pixel array. Must be one-dimensional, monotonically
            increasing, and have `row_array.size = value_array.shape[0]`.
        col_array : numpy.ndarray
            The columns of the pixel array. Must be one-dimensional, monotonically
            increasing, and have and have `col_array.size = value_array.shape[1]`.
        value_array : numpy.ndarray
            The values array, whihc must be two or three dimensional. If this has
            complex dtype and `complex_valued=False`, then the :func:`numpy.abs`
            will be applied.
        value_is_flat : bool
            If `True`, then `value_array` must be exactly two-dimensional.

        Returns
        -------
        numpy.ndarray
        """

        # verify numpy arrays
        if not isinstance(value_array, numpy.ndarray):
            raise TypeError('value_array must be numpy.ndarray, got type {}'.format(type(value_array)))
        if not isinstance(row_array, numpy.ndarray):
            raise TypeError('row_array must be numpy.ndarray, got type {}'.format(type(row_array)))
        if not isinstance(col_array, numpy.ndarray):
            raise TypeError('col_array must be numpy.ndarray, got type {}'.format(type(col_array)))

        # verify array shapes make sense
        if value_array.ndim not in (2, 3):
            raise ValueError('value_array must be two or three dimensional')
        if row_array.ndim != 1 or row_array.size != value_array.shape[0]:
            raise ValueError(
                'We must have row_array is one dimensional and row_array.size = value.array.shape[0]. '
                'Got row_array.shape = {}, and value_array = {}'.format(row_array.shape, value_array.shape))
        if col_array.ndim != 1 or col_array.size != value_array.shape[1]:
            raise ValueError(
                'We must have col_array is one dimensional and col_array.size = value.array.shape[1]. '
                'Got col_array.shape = {}, and value_array = {}'.format(col_array.shape, value_array.shape))
        if value_is_flat and len(value_array.shape) != 2:
            raise ValueError('value_array must be two-dimensional. Got shape {}'.format(value_array.shape))

        # verify row_array/col_array are monotonically increasing
        if numpy.any(numpy.diff(row_array.astype('float64')) <= 0):
            raise ValueError('row_array must be monotonically increasing.')
        if numpy.any(numpy.diff(col_array.astype('float64')) <= 0):
            raise ValueError('col_array must be monotonically increasing.')

        # address the dtype of value_array
        if (not self._complex_valued) and numpy.iscomplexobj(value_array):
            return numpy.abs(value_array)

        return value_array

    def get_orthorectified_from_array(self, ortho_bounds, row_array, col_array, value_array):
        """
        Construct the orthorectified array covering the orthorectified region given by
        `ortho_bounds` based on the `values_array`, which spans the pixel region defined by
        `row_array` and `col_array`.

        This is mainly a helper method, and should only be called directly for specific and
        directed reasons.

        Parameters
        ----------
        ortho_bounds : numpy.ndarray
            Determines the orthorectified bounds region, of the form
            `(min row, max row, min column, max column)`.
        row_array : numpy.ndarray
            The rows of the pixel array. Must be one-dimensional, monotonically increasing,
            and have `row_array.size = value_array.shape[0]`.
        col_array
            The columns of the pixel array. Must be one-dimensional, monotonically increasing,
            and have `col_array.size = value_array.shape[1]`.
        value_array
            The values array. If this has complex dtype and `complex_valued=False`, then
            the :func:`numpy.abs` will be applied.

        Returns
        -------
        numpy.ndarray
        """

        # validate our inputs
        value_array = self._validate_row_col_values(row_array, col_array, value_array, value_is_flat=False)
        if value_array.size == 0:
            if value_array.ndim == 3:
                return self._initialize_workspace(ortho_bounds, final_dimension=value_array.shape[2])
            else:
                return self._initialize_workspace(ortho_bounds)
        if value_array.ndim == 2:
            return self._get_orthrectified_from_array_flat(ortho_bounds, row_array, col_array, value_array)
        else:  # it must be three dimensional, as checked by _validate_row_col_values()
            ortho_array = self._initialize_workspace(ortho_bounds, final_dimension=value_array.shape[2])
            for i in range(value_array.shape[2]):
                ortho_array[:, :, i] = self._get_orthrectified_from_array_flat(
                    ortho_bounds, row_array, col_array, value_array[:, :, i])
            return ortho_array

    def get_orthorectified_for_ortho_bounds(self, bounds):
        """
        Determine the array corresponding to the array of bounds given in
        ortho-rectified pixel coordinates.

        Parameters
        ----------
        bounds : numpy.ndarray|list|tuple
            Of the form `(row_min, row_max, col_min, col_max)`. Note that non-integer
            values will be expanded outwards (floor of minimum and ceil at maximum).
            Following Python convention, this will be inclusive at the minimum and
            exclusive at the maximum.

        Returns
        -------
        numpy.ndarray
        """

        ortho_bounds, nominal_pixel_bounds = self.extract_pixel_bounds(bounds)
        # extract the values - ensure that things are within proper image bounds
        pixel_bounds = self.get_real_pixel_bounds(nominal_pixel_bounds)
        pixel_array = self.reader[
            pixel_bounds[0]:pixel_bounds[1], pixel_bounds[2]:pixel_bounds[3], self.index]
        row_arr = numpy.arange(pixel_bounds[0], pixel_bounds[1])
        col_arr = numpy.arange(pixel_bounds[2], pixel_bounds[3])
        return self.get_orthorectified_from_array(ortho_bounds, row_arr, col_arr, pixel_array)

    def get_orthorectified_for_pixel_bounds(self, pixel_bounds):
        """
        Determine the array corresponding to the given array bounds given in reader
        pixel coordinates.

        Parameters
        ----------
        pixel_bounds : numpy.ndarray|list|tuple
            Of the form `(row_min, row_max, col_min, col_max)`.

        Returns
        -------
        numpy.ndarray
        """

        pixel_bounds, pixel_rect = self.bounds_to_rectangle(pixel_bounds)
        return self.get_orthorectified_for_pixel_object(pixel_rect)

    def get_orthorectified_for_pixel_object(self, coordinates):
        """
        Determine the ortho-rectified rectangular array values, which will bound
        the given object - with coordinates expressed in pixel space.

        Parameters
        ----------
        coordinates : GeometryObject|numpy.ndarray|list|tuple
            The coordinate system of the input will be assumed to be pixel space.

        Returns
        -------
        numpy.ndarray
        """

        bounds = self.get_orthorectification_bounds_from_pixel_object(coordinates)
        return self.get_orthorectified_for_ortho_bounds(bounds)

    def get_orthorectified_for_latlon_object(self, ll_coordinates):
        """
        Determine the ortho-rectified rectangular array values, which will bound
        the given object - with coordinates expressed in lat/lon space.

        Parameters
        ----------
        ll_coordinates : GeometryObject|numpy.ndarray|list|tuple
            The coordinate system of the input will be assumed to be pixel space.
            **Note** a GeometryObject is expected to follow lon/lat ordering paradigm,
            by convention.

        Returns
        -------
        numpy.ndarray
        """

        bounds = self.get_orthorectification_bounds_from_latlon_object(ll_coordinates)
        return self.get_orthorectified_for_ortho_bounds(bounds)

    def _setup_flat_workspace(self, ortho_bounds, row_array, col_array, value_array):
        """
        Helper method for setting up the flat workspace.

        Parameters
        ----------
        ortho_bounds : numpy.ndarray
            Determines the orthorectified bounds region, of the form
            `(min row, max row, min column, max column)`.
        row_array : numpy.ndarray
            The rows of the pixel array. Must be one-dimensional, monotonically
            increasing, and have `row_array.size = value_array.shape[0]`.
        col_array : numpy.ndarray
            The columns of the pixel array. Must be one-dimensional, monotonically
            increasing, and have `col_array.size = value_array.shape[1]`.
        value_array : numpy.ndarray
            The values array. If this has complex dtype and `complex_valued=False`,
            then the :func:`numpy.abs` will be applied.

        Returns
        -------
        (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray)
        """

        value_array = self._validate_row_col_values(row_array, col_array, value_array, value_is_flat=True)
        # set up the results workspace
        ortho_array = self._initialize_workspace(ortho_bounds)
        # determine the pixel coordinates for the ortho coordinates meshgrid
        ortho_mesh = self._get_ortho_mesh(ortho_bounds)
        # determine the nearest neighbor pixel coordinates
        pixel_mesh = self.proj_helper.ortho_to_pixel(ortho_mesh)
        pixel_rows = pixel_mesh[:, :, 0]
        pixel_cols = pixel_mesh[:, :, 1]
        return value_array, pixel_rows, pixel_cols, ortho_array

    def _get_orthrectified_from_array_flat(self, ortho_bounds, row_array, col_array, value_array):
        """
        Construct the orthorecitified array covering the orthorectified region given by
        `ortho_bounds` based on the `values_array`, which spans the pixel region defined by
        `row_array` and `col_array`.

        Parameters
        ----------
        ortho_bounds : numpy.ndarray
            Determines the orthorectified bounds region, of the form
            `(min row, max row, min column, max column)`.
        row_array : numpy.ndarray
            The rows of the pixel array. Must be one-dimensional, monotonically
            increasing, and have `row_array.size = value_array.shape[0]`.
        col_array
            The columns of the pixel array. Must be one-dimensional, monotonically
            increasing, and have `col_array.size = value_array.shape[1]`.
        value_array
            The values array. If this has complex dtype and `complex_valued=False`,
            then the :func:`numpy.abs` will be applied.

        Returns
        -------
        numpy.ndarray
        """

        raise NotImplementedError


class NearestNeighborMethod(OrthorectificationHelper):
    """
    Nearest neighbor ortho-rectification method.

    .. warning::
        Modification of the proj_helper parameters when the default full image
        bounds have been defained (i.e. sicd.RadarCollection.Area is defined) may
        result in unintended results.
    """

    def __init__(self, reader, index=0, proj_helper=None, complex_valued=False,
                 pad_value=None, apply_radiometric=None, subtract_radiometric_noise=False):
        """

        Parameters
        ----------
        reader : SICDTypeReader
        index : int
        proj_helper : None|ProjectionHelper
            If `None`, this will default to `PGRatPolyProjection(<sicd>)` unless there is
            a SarpyRatPolyError, when it will fall back to `PGProjection(<sicd>)`,
            where `<sicd>` will be the sicd from `reader` at `index`. Otherwise,
            it is the user's responsibility to ensure that `reader`, `index` and
            `proj_helper` are in sync.
        complex_valued : bool
            Do we want complex values returned? If `False`, the magnitude values
            will be used.
        pad_value : None|Any
            Value to use for any out-of-range pixels. Defaults to `0` if not provided.
        apply_radiometric : None|str
            **Only valid if `complex_valued=False`**. If provided, must be one of
            `['RCS', 'Sigma0', 'Gamma0', 'Beta0']` (not case-sensitive). This will
            apply the given radiometric scale factor to the array values.
        subtract_radiometric_noise : bool
            **Only has any effect if `apply_radiometric` is provided.** This indicates that
            the radiometric noise should be subtracted prior to applying the given
            radiometric scale factor.
        """

        super(NearestNeighborMethod, self).__init__(
            reader, index=index, proj_helper=proj_helper, complex_valued=complex_valued,
            pad_value=pad_value, apply_radiometric=apply_radiometric,
            subtract_radiometric_noise=subtract_radiometric_noise)

    def _get_orthrectified_from_array_flat(self, ortho_bounds, row_array, col_array, value_array):
        # setup the result workspace
        value_array, pixel_rows, pixel_cols, ortho_array = self._setup_flat_workspace(
            ortho_bounds, row_array, col_array, value_array)
        # potentially apply the radiometric parameters to the value array
        value_array = self._apply_radiometric_params(row_array, col_array, value_array)
        if value_array.size > 0:
            # determine the in bounds points
            mask = self._get_mask(pixel_rows, pixel_cols, row_array, col_array)
            # determine the nearest neighbors for our row/column indices
            row_inds = numpy.digitize(pixel_rows[mask], row_array)
            col_inds = numpy.digitize(pixel_cols[mask], col_array)
            ortho_array[mask] = value_array[row_inds, col_inds]
        return ortho_array


class BivariateSplineMethod(OrthorectificationHelper):
    """
    Bivariate spline interpolation ortho-rectification method.

    .. warning::
        Modification of the proj_helper parameters when the default full image
        bounds have been defained (i.e. sicd.RadarCollection.Area is defined) may
        result in unintended results.
    """

    __slots__ = ('_row_order', '_col_order')

    def __init__(self, reader, index=0, proj_helper=None, complex_valued=False,
                 pad_value=None, apply_radiometric=None, subtract_radiometric_noise=False,
                 row_order=1, col_order=1):
        """

        Parameters
        ----------
        reader : SICDTypeReader
        index : int
        proj_helper : None|ProjectionHelper
            If `None`, this will default to `PGRatPolyProjection(<sicd>)` unless there is
            a SarpyRatPolyError, when it will fall back to `PGProjection(<sicd>)`,
            where `<sicd>` will be the sicd from `reader` at `index`. Otherwise,
            it is the user's responsibility to ensure that `reader`, `index` and
            `proj_helper` are in sync.
        complex_valued : bool
            Do we want complex values returned? If `False`, the magnitude values
            will be used.
        pad_value : None|Any
            Value to use for any out-of-range pixels. Defaults to `0` if not provided.
        apply_radiometric : None|str
            **Only valid if `complex_valued=False`**. If provided, must be one of
            `['RCS', 'Sigma0', 'Gamma0', 'Beta0']` (not case-sensitive). This will
            apply the given radiometric scale factor to the array values.
        subtract_radiometric_noise : bool
            **Only has any effect if `apply_radiometric` is provided.** This indicates that
            the radiometric noise should be subtracted prior to applying the given
            radiometric scale factor.
        row_order : int
            The row degree for the spline.
        col_order : int
            The column degree for the spline.
        """

        self._row_order = None
        self._col_order = None
        if complex_valued:
            raise ValueError('BivariateSpline only supports real valued results for now.')
        super(BivariateSplineMethod, self).__init__(
            reader, index=index, proj_helper=proj_helper, complex_valued=complex_valued,
            pad_value=pad_value, apply_radiometric=apply_radiometric,
            subtract_radiometric_noise=subtract_radiometric_noise)
        self.row_order = row_order
        self.col_order = col_order

    @property
    def row_order(self):
        """
        int : The spline order for the x/row coordinate, where `1 <= row_order <= 5`.
        """
        return self._row_order

    @row_order.setter
    def row_order(self, value):
        value = int(value)
        if not (1 <= value <= 5):
            raise ValueError('row_order must take value between 1 and 5.')
        self._row_order = value

    @property
    def col_order(self):
        """
        int : The spline order for the y/col coordinate, where `1 <= col_order <= 5`.
        """

        return self._col_order

    @col_order.setter
    def col_order(self, value):
        value = int(value)
        if not (1 <= value <= 5):
            raise ValueError('col_order must take value between 1 and 5.')
        self._col_order = value

    def _get_orthrectified_from_array_flat(self, ortho_bounds, row_array, col_array, value_array):
        # setup the result workspace
        value_array, pixel_rows, pixel_cols, ortho_array = self._setup_flat_workspace(
            ortho_bounds, row_array, col_array, value_array)
        value_array = self._apply_radiometric_params(row_array, col_array, value_array)

        if value_array.size > 0:
            # set up our spline
            sp = RectBivariateSpline(row_array, col_array, value_array, kx=self.row_order, ky=self.col_order, s=0)
            # determine the in bounds points
            mask = self._get_mask(pixel_rows, pixel_cols, row_array, col_array)
            result = sp.ev(pixel_rows[mask], pixel_cols[mask])
            # potentially apply the radiometric parameters
            ortho_array[mask] = result
        return ortho_array



"""
This module provides tools for creating kmz products for a CPHD type element.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Valkyrie Systems Corporation"

import logging
import os

import numpy as np
import scipy.constants
import scipy.optimize
import shapely.geometry as shg

import sarpy.visualization.kmz_product_creation as kpc
from sarpy.io.kml import Document
from sarpy.visualization import kmz_utils

logger = logging.getLogger(__name__)


def _create_cphd_styles(kmz_document):
    kmz_utils.add_antenna_styles(kmz_document)

    # iarp - intended for basic point clamped to ground
    label = {"color": "ff50c0c0", "scale": "1.0"}
    icon = {
        "color": "ff5050c0",
        "scale": "1.5",
        "icon_ref": "http://maps.google.com/mapfiles/kml/shapes/shaded_dot.png",
    }
    kmz_document.add_style("iarp_high", label_style=label, icon_style=icon)
    label["scale"] = "0.75"
    icon["scale"] = "1.0"
    kmz_document.add_style("iarp_low", label_style=label, icon_style=icon)
    kmz_document.add_style_map("iarp", "iarp_high", "iarp_low")

    # srp position style - intended for gx track
    line = {"color": "ff50ff50", "width": "1.5"}
    label = {"color": "ffc0c0c0", "scale": "0.5"}
    icon = {
        "scale": "2.0",
        "icon_ref": "http://maps.google.com/mapfiles/kml/shapes/placemark_circle.png",
    }
    poly = {"color": "a050ff50"}
    kmz_document.add_style(
        "srp_high", line_style=line, label_style=label, icon_style=icon, poly_style=poly
    )
    line["width"] = "1.0"
    icon["scale"] = "1.0"
    poly = {"color": "7050ff50"}
    kmz_document.add_style(
        "srp_low", line_style=line, label_style=label, icon_style=icon, poly_style=poly
    )
    kmz_document.add_style_map("srp", "srp_high", "srp_low")

    kmz_utils.add_polygon_style(
        kmz_document,
        "globalimagearea",
        bbggrr="aa00ff",
        low_aa="00",
        low_width="1.0",
        high_aa="00",
        high_width="1.5",
    )

    kmz_utils.add_polygon_style(
        kmz_document,
        "channelimagearea",
        bbggrr="00aa7f",
        low_aa="70",
        low_width="1.0",
        high_aa="a0",
        high_width="1.5",
    )

    def _add_toa_style(name, aabbggrr):
        kmz_document.add_style(
            name + "_high",
            line_style={"color": aabbggrr, "width": "1.5"},
            poly_style={"color": aabbggrr},
        )
        kmz_document.add_style(
            name + "_low",
            line_style={"color": aabbggrr, "width": "1.0"},
            poly_style={"color": aabbggrr},
        )
        kmz_document.add_style_map(name, name + "_high", name + "_low")

    _add_toa_style("toa", "ff7a40ec")
    _add_toa_style("toae", "ffffee58")


def imagearea_kml_coord(scenecoords_node, imagearea_node):
    """Create KML coords of an imagearea footprint"""
    if imagearea_node.Polygon is not None:
        verts = [
            vert.get_array()
            for vert in sorted(imagearea_node.Polygon, key=lambda x: x.index)
        ]
    else:
        x1, y1 = imagearea_node.X1Y1.get_array()
        x2, y2 = imagearea_node.X2Y2.get_array()
        verts = [(x1, y1), (x1, y2), (x2, y2), (x2, y1)]
    verts.append(verts[0])
    return _xy_to_kml_coord(scenecoords_node, verts)


def cphd_create_kmz_view(reader, output_directory, file_stem="view"):
    """
    Create a kmz view for the reader contents.

    Parameters
    ----------
    reader : CPHDTypeReader
    output_directory : str
    file_stem : str

    Returns
    -------
    None

    Examples
    --------
    .. code-block:: python

        import logging
        logger = logging.getLogger('sarpy')
        logger.setLevel('INFO')

        import os
        import sarpy.io.phase_history
        from sarpy.visualization.cphd_kmz_product_creation import cphd_create_kmz_view

        test_root = '<root directory>'
        reader = sarpy.io.phase_history.open(os.path.join(test_root, '<file name>>'))
        cphd_create_kmz_view(reader, test_root,
                             file_stem='View-<something descriptive>')

    """

    def add_global(kmz_doc, root):
        logger.info("Adding global to kmz.")
        folder = kmz_doc.add_container(
            par=root, the_type="Folder", name="Global", description="Global Metadata"
        )
        iarp_llh = reader.cphd_meta.SceneCoordinates.IARP.LLH.get_array()
        frm = "{1:0.8f},{0:0.8f},0"  # SICD version throws away height
        coords = frm.format(*iarp_llh)
        placemark = kmz_doc.add_container(
            par=folder, name="IARP", description="IARP", styleUrl="#iarp"
        )
        kmz_doc.add_point(coords, par=placemark, altitudeMode="absolute")

        ia_coords = imagearea_kml_coord(
            reader.cphd_meta.SceneCoordinates,
            reader.cphd_meta.SceneCoordinates.ImageArea,
        )
        placemark = kmz_doc.add_container(
            par=folder,
            name="ImageArea",
            description="ImageArea",
            styleUrl="#globalimagearea",
        )
        kmz_doc.add_polygon(
            " ".join(ia_coords),
            par=placemark,
            name="ImageArea",
            altitudeMode="absolute",
        )

    def add_channel(kmz_doc, root, channel_name):
        channel_names = [chan.Identifier for chan in reader.cphd_meta.Data.Channels]
        channel_index = channel_names.index(channel_name)
        pvp_array = reader.read_pvp_array(channel_index)
        logger.info(f"Adding channel '{channel_name}' to kmz.")

        if "SIGNAL" in pvp_array.dtype.names:
            signal = pvp_array["SIGNAL"]
        else:
            num_vectors = reader.cphd_meta.Data.Channels[channel_index].NumVectors
            signal = np.ones(num_vectors)

        num_subselect = 24
        indices = np.where(signal == 1)[0]
        geom_indices = indices[
            np.round(
                np.linspace(0, indices.size - 1, num_subselect, endpoint=True)
            ).astype(int)
        ]

        times = (
            pvp_array["TxTime"][geom_indices] + pvp_array["RcvTime"][geom_indices]
        ) / 2.0
        arp_pos = (
            pvp_array["TxPos"][geom_indices] + pvp_array["RcvPos"][geom_indices]
        ) / 2.0
        srp_pos = pvp_array["SRPPos"][geom_indices]
        collection_start = reader.cphd_meta.Global.Timeline.CollectionStart.astype(
            "datetime64[us]"
        )
        whens = collection_start + (times * 1e6).astype("timedelta64[us]")
        whens = [str(time) + "Z" for time in whens]

        time_args = {"beginTime": whens[0], "endTime": whens[-1]}

        folder = kmz_doc.add_container(
            par=root,
            the_type="Folder",
            name=f"Channel {channel_name}",
            description=f"Channel {channel_name}",
            when=whens[0],
        )
        arp_coords = kmz_utils.ecef_to_kml_coord(arp_pos)
        placemark = kmz_doc.add_container(
            par=folder,
            name=channel_name,
            description=f"aperture position for channel {channel_name}",
            styleUrl="#arp",
            **time_args,
        )
        kmz_doc.add_gx_track(
            arp_coords,
            whens,
            par=placemark,
            extrude=True,
            tesselate=True,
            altitudeMode="absolute",
        )

        srp_coords = kmz_utils.ecef_to_kml_coord(srp_pos)
        placemark = kmz_doc.add_container(
            par=folder,
            name="SRP",
            description=f"stabilization reference point for channel {channel_name}",
            styleUrl="#srp",
            **time_args,
        )
        kmz_doc.add_gx_track(
            srp_coords,
            whens,
            par=placemark,
            extrude=True,
            tesselate=True,
            altitudeMode="absolute",
        )

        if reader.cphd_meta.Channel.Parameters[channel_index].ImageArea is not None:
            ia_coords = imagearea_kml_coord(
                reader.cphd_meta.SceneCoordinates,
                reader.cphd_meta.Channel.Parameters[channel_index].ImageArea,
            )
            placemark = kmz_doc.add_container(
                par=folder,
                name="ImageArea",
                description=f"ImageArea for channel {channel_name}",
                styleUrl="#channelimagearea",
            )
            kmz_doc.add_polygon(
                " ".join(ia_coords),
                par=placemark,
                name=f"ImageArea for channel {channel_name}",
                altitudeMode="absolute",
            )

        chan_params = reader.cphd_meta.Channel.Parameters[channel_index]

        antenna_folder = kmz_doc.add_container(
            the_type="Folder",
            par=folder,
            name="Antenna",
            description=f"Antenna Aiming for channel {channel_name}",
        )
        boresight_folder = kmz_doc.add_container(
            the_type="Folder",
            par=antenna_folder,
            name="Boresights",
            description=f"Boresights for channel {channel_name}",
        )
        footprint_folder = kmz_doc.add_container(
            the_type="Folder",
            par=antenna_folder,
            name="-3dB Footprints",
            description=f"Beam Footprints for channel {channel_name}",
        )

        footprint_labels = {
            "start": indices[0],
            "middle": indices[len(indices) // 2],
            "end": indices[-1],
        }

        footprints = {}
        for txrcv in ("Tx", "Rcv"):
            aiming = antenna_aiming(
                reader.cphd_meta.Antenna,
                pvp_array,
                txrcv=txrcv,
                apc_id=getattr(chan_params.Antenna, f"{txrcv}APCId"),
                antpat_id=getattr(chan_params.Antenna, f"{txrcv}APATId"),
            )
            if not aiming:
                break

            for boresight_type in ("mechanical", "electrical"):
                visibility = txrcv == "Rcv"  # only display Rcv by default
                name = f"{txrcv} {boresight_type} boresight"

                on_earth_ecf = np.asarray(
                    [
                        kmz_utils.ray_intersect_earth(apc_pos, along)
                        for apc_pos, along in zip(
                            aiming["raw"]["positions"][geom_indices],
                            aiming[boresight_type][geom_indices],
                        )
                    ]
                )

                placemark = kmz_doc.add_container(
                    par=boresight_folder,
                    name=name,
                    description=f"{name} for channel {channel_name}<br><br>Highlighted edge indicates start time",
                    styleUrl=f"#{boresight_type}_boresight",
                    visibility=visibility,
                )
                boresight_coords = kmz_utils.ecef_to_kml_coord(on_earth_ecf)
                kmz_utils.add_los_polygon(
                    kmz_doc, placemark, arp_coords, boresight_coords
                )

            array_gain_poly = aiming["raw"]["pattern"].Array.GainPoly
            element_gain_poly = aiming["raw"]["pattern"].Element.GainPoly
            footprints[txrcv] = kmz_utils.make_beam_footprints(
                aiming,
                footprint_labels,
                array_gain_poly,
                element_gain_poly,
                contour_level=-3,
            )
            for when, this_footprint in footprints[txrcv].items():
                name = f"{txrcv} beam footprint @ {when}"
                timestamp = (
                    str(
                        collection_start
                        + (this_footprint["time"] * 1e6).astype("timedelta64[us]")
                    )
                    + "Z"
                )
                placemark = kmz_doc.add_container(
                    par=footprint_folder,
                    name=name,
                    description=f"{name} for channel {channel_name}",
                    styleUrl=f"#{txrcv.lower()}_beam_footprint",
                    visibility=True,
                    when=timestamp,
                )
                coords = kmz_utils.ecef_to_kml_coord(this_footprint["contour"])
                kmz_doc.add_polygon(
                    " ".join(coords),
                    par=placemark,
                )

        # Try to add TOA extents
        if reader.cphd_meta.SceneCoordinates.ReferenceSurface.Planar is None:
            return

        iarp_ecf = reader.cphd_meta.SceneCoordinates.IARP.ECF.get_array()
        uiax = (
            reader.cphd_meta.SceneCoordinates.ReferenceSurface.Planar.uIAX.get_array()
        )
        uiay = (
            reader.cphd_meta.SceneCoordinates.ReferenceSurface.Planar.uIAY.get_array()
        )

        def _in_scene_plane(x):
            return np.squeeze(np.atleast_2d(x) @ np.stack((uiax, uiay), axis=-1))

        valid_toa_points = {}
        for when, index in footprint_labels.items():
            # define "bp" coordinate system aligned with bistatic pointing vector in scene plane
            bp = (
                _unit(pvp_array["TxPos"][index] - pvp_array["SRPPos"][index])
                + _unit(pvp_array["RcvPos"][index] - pvp_array["SRPPos"][index])
            ) / 2.0
            bp_in_plane = _unit(_in_scene_plane(bp))
            cross_bp_in_plane = np.array([-bp_in_plane[1], bp_in_plane[0]])
            srp_in_plane = _in_scene_plane(pvp_array["SRPPos"][index] - iarp_ecf)
            transform_from_bp_to_scene = np.vstack(
                (
                    np.stack([bp_in_plane, cross_bp_in_plane, srp_in_plane], axis=-1),
                    np.array([0, 0, 1]),
                )
            )
            transform_from_scene_to_bp = np.linalg.inv(transform_from_bp_to_scene)

            def _bp_to_ecf(bp):
                xy = _apply_homogeneous_transform(bp, transform_from_bp_to_scene)
                return _xy_to_ecf(reader.cphd_meta.SceneCoordinates, xy.T)

            def _get_toa_point_ecf(initial_bp, toa_target):
                def _errfunc(offset):
                    pt_ecf = _bp_to_ecf(
                        initial_bp[:, np.newaxis]
                        + np.vstack([offset, np.zeros(len(offset))])
                    )
                    this_toa = _geom_to_toa(
                        pt_ecf,
                        pvp_array["TxPos"][index],
                        pvp_array["RcvPos"][index],
                        pvp_array["SRPPos"][index],
                    )
                    return np.abs(this_toa - toa_target)

                res = scipy.optimize.minimize(_errfunc, 0, method="Nelder-Mead")
                if res.success:
                    return _bp_to_ecf((initial_bp + [res.x.item(), 0])[:, np.newaxis])
                return np.full((1, 3), np.nan)

            # Try to use Tx and Rcv contours to determine location of TOA extent lines
            toa_point_center = None
            toa_point_width = None
            tx_footprint = footprints.get("Tx", {}).get(when)
            rcv_footprint = footprints.get("Rcv", {}).get(when)
            if tx_footprint is not None and rcv_footprint is not None:
                tx_contour_imcoords = shg.MultiPoint(
                    _in_scene_plane(np.array(tx_footprint["contour"]) - iarp_ecf)
                ).convex_hull
                rcv_contour_imcoords = shg.MultiPoint(
                    _in_scene_plane(np.array(rcv_footprint["contour"]) - iarp_ecf)
                ).convex_hull
                if tx_contour_imcoords.intersects(rcv_contour_imcoords):
                    combined_contour = tx_contour_imcoords | rcv_contour_imcoords
                    combined_contour_in_bp = shg.Polygon(
                        _apply_homogeneous_transform(
                            np.array(combined_contour.exterior.coords.xy),
                            transform_from_scene_to_bp,
                        ).T
                    )
                    toa_point_center = np.array(
                        [
                            combined_contour_in_bp.centroid.x,
                            combined_contour_in_bp.centroid.y,
                        ]
                    )
                    toa_point_width = (
                        combined_contour_in_bp.bounds[3]
                        - combined_contour_in_bp.bounds[1]
                    )

            # Fallback to the SRP for center and TOA extent for width if the location
            # could not be determined from the contours
            toa_point_center = (
                np.zeros(2) if toa_point_center is None else toa_point_center
            )
            centers = {
                x: _get_toa_point_ecf(toa_point_center, pvp_array[x][index])
                for x in ("TOA1", "TOA2")
            }

            toa_point_width = toa_point_width or np.linalg.norm(
                centers["TOA2"] - centers["TOA1"]
            )
            toa_points = {
                x: np.concatenate(
                    [
                        _get_toa_point_ecf(
                            toa_point_center + [0, toa_point_width / 2],
                            pvp_array[x][index],
                        ),
                        centers[x],
                        _get_toa_point_ecf(
                            toa_point_center + [0, -toa_point_width / 2],
                            pvp_array[x][index],
                        ),
                    ],
                    axis=0,
                )
                for x in ("TOA1", "TOA2")
            }
            if {"TOAE1", "TOAE2"}.issubset(pvp_array.dtype.names):
                for param in ("TOAE1", "TOAE2"):
                    toa_points[param] = np.concatenate(
                        [
                            _get_toa_point_ecf(
                                toa_point_center + [0, offset], pvp_array[param][index]
                            )
                            for offset in (toa_point_width / 2, 0, -toa_point_width / 2)
                        ],
                        axis=0,
                    )
            valid_toa_points[when] = {
                k: v for k, v in toa_points.items() if np.isfinite(v).all()
            }

        if valid_toa_points:
            toa_folder = kmz_doc.add_container(
                par=folder,
                the_type="Folder",
                name="TOA extents",
                description=f"TOA extents for channel {channel_name}",
            )
            for when, toa_points in valid_toa_points.items():
                for label, ecf_points in toa_points.items():
                    name = f"{label} @ {when}"
                    placemark = kmz_doc.add_container(
                        par=toa_folder,
                        name=name,
                        description=f"{name} for channel {channel_name}",
                        styleUrl=f"#{label[:-1].lower()}",
                    )
                    kmz_doc.add_line_string(
                        coords=" ".join(kmz_utils.ecef_to_kml_coord(ecf_points)),
                        par=placemark,
                        altitudeMode="absolute",
                        extrude=True,
                    )

    kmz_file = os.path.join(output_directory, f"{file_stem}_cphd.kmz")
    with prepare_kmz_file(kmz_file, name=reader.file_name) as kmz_doc:
        root = kmz_doc.add_container(
            the_type="Folder", name=reader.cphd_meta.CollectionID.CoreName
        )
        add_global(kmz_doc, root)
        for chan in reader.cphd_meta.Data.Channels:
            add_channel(kmz_doc, root, channel_name=chan.Identifier)


def _apply_homogeneous_transform(x, t, is_position=True):
    homogeneous_coord = np.ones if is_position else np.zeros
    return (t @ np.vstack([x, homogeneous_coord((1, x.shape[1]))]))[:-1, ...]


def _geom_to_toa(ecef_point, tx_apc_pos, rcv_apc_pos, scp):
    """
    Calculate the time for signal to go from `tx_apc_pos` to `ecef_point` to `rcv_apc_pos` relative to SCP.
    """

    def one_dir(apc_pos, pt):
        los_to_apc = apc_pos - pt
        toa = np.linalg.norm(los_to_apc, axis=-1) / scipy.constants.speed_of_light
        return toa

    def one_dir_rel(apc_pos):
        pt_toa = one_dir(apc_pos, ecef_point)
        scp_toa = one_dir(apc_pos, scp)
        return pt_toa - scp_toa

    tx_toa = one_dir_rel(tx_apc_pos)
    rcv_toa = one_dir_rel(rcv_apc_pos)
    return tx_toa + rcv_toa


def _unit(vec, axis=-1):
    return vec / np.linalg.norm(vec, axis=axis, keepdims=True)


def prepare_kmz_file(file_name, **args):
    """
    Prepare a kmz document and archive for exporting.

    Parameters
    ----------
    file_name : str
    args
        Passed through to the Document constructor.

    Returns
    -------
    Document
    """

    document = Document(file_name=file_name, **args)
    kpc._create_sicd_styles(document)
    _create_cphd_styles(document)
    return document


def _xy_to_ecf(scenecoords_node, xy):
    """Convert a ReferenceSurface XY location to ECEF coordinates"""
    xy = np.atleast_2d(xy)
    if scenecoords_node.ReferenceSurface.Planar is not None:
        iarp_ecf = scenecoords_node.IARP.ECF.get_array()
        uiax = scenecoords_node.ReferenceSurface.Planar.uIAX.get_array()
        uiay = scenecoords_node.ReferenceSurface.Planar.uIAY.get_array()
        xy_ecf = iarp_ecf + xy[:, 0, np.newaxis] * uiax + xy[:, 1, np.newaxis] * uiay
    else:
        raise NotImplementedError
    return xy_ecf


def _xy_to_kml_coord(scenecoords_node, xy):
    """Convert a ReferenceSurface XY location to a kml coordinate"""
    xy_ecf = _xy_to_ecf(scenecoords_node, xy)
    return kmz_utils.ecef_to_kml_coord(xy_ecf)


def antenna_aiming(antenna_node, pvp_array, *, txrcv, apc_id, antpat_id):
    """Compile antenna aiming metadata"""
    if antenna_node is None:
        return {}

    apcs = {apc.Identifier: apc for apc in antenna_node.AntPhaseCenter}
    acfs = {acf.Identifier: acf for acf in antenna_node.AntCoordFrame}
    patterns = {antpat.Identifier: antpat for antpat in antenna_node.AntPattern}

    positions = pvp_array[f"{txrcv}Pos"]
    times = pvp_array[f"{txrcv}Time"]
    if {f"{txrcv}AC{d}" for d in "XY"}.issubset(pvp_array.dtype.names):
        uacx = pvp_array[f"{txrcv}ACX"]
        uacy = pvp_array[f"{txrcv}ACY"]
    else:
        acf_id = apcs[apc_id].ACFId
        uacx = acfs[acf_id].XAxisPoly(times)
        uacy = acfs[acf_id].YAxisPoly(times)
    uacz = np.cross(uacx, uacy)

    pointing = {
        "antpat_id": antpat_id,
        "raw": {
            "positions": positions,
            "times": times,
            "uacx": uacx,
            "uacy": uacy,
            "uacz": uacz,
            "pattern": patterns[antpat_id],
        },
        "mechanical": uacz,
    }

    if f"{txrcv}EB" in pvp_array.dtype.names:
        ebpvp = pvp_array[f"{txrcv}EB"]
        eb_dcx = ebpvp[:, 0]
        eb_dcy = ebpvp[:, 1]
    else:
        eb_dcx = patterns[antpat_id].EB.DCXPoly(times)
        eb_dcy = patterns[antpat_id].EB.DCYPoly(times)

    pointing["raw"]["eb_dcx"] = eb_dcx
    pointing["raw"]["eb_dcy"] = eb_dcy

    pointing["electrical"] = kmz_utils.acf_to_ecef(eb_dcx, eb_dcy, uacx, uacy)

    return pointing



"""
The graphics header element definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from .base import NITFElement, UserHeaderType, _IntegerDescriptor,\
    _StringDescriptor, _StringEnumDescriptor, _NITFElementDescriptor
from .security import NITFSecurityTags


class GraphicsSegmentHeader(NITFElement):
    """
    Graphics segment subheader - see standards document Joint BIIF Profile (JBP) for more
    information.
    """
    _ordering = (
        'SY', 'SID', 'SNAME', 'Security', 'ENCRYP', 'SFMT',
        'SSTRUCT', 'SDLVL', 'SALVL', 'SLOC', 'SBND1',
        'SCOLOR', 'SBND2', 'SRES2', 'UserHeader')
    _lengths = {
        'SY': 2, 'SID': 10, 'SNAME': 20, 'ENCRYP': 1,
        'SFMT': 1, 'SSTRUCT': 13, 'SDLVL': 3, 'SALVL': 3,
        'SLOC': 10, 'SBND1': 10, 'SCOLOR': 1, 'SBND2': 10,
        'SRES2': 2}
    SY = _StringEnumDescriptor(
        'SY', True, 2, {'SY', }, default_value='SY',
        docstring='File part type.')  # type: str
    SID = _StringDescriptor(
        'SID', True, 10, default_value='',
        docstring='Graphic Identifier. This field shall contain a valid alphanumeric identification code '
                  'associated with the graphic. The valid codes are determined by the application.')  # type: str
    SNAME = _StringDescriptor(
        'SNAME', True, 20, default_value='',
        docstring='Graphic name. This field shall contain an alphanumeric name for the graphic.')  # type: str
    Security = _NITFElementDescriptor(
        'Security', True, NITFSecurityTags, default_args={},
        docstring='The security tags.')  # type: NITFSecurityTags
    ENCRYP = _StringEnumDescriptor(
        'ENCRYP', True, 1, {'0'}, default_value='0',
        docstring='Encryption.')  # type: str
    SFMT = _StringDescriptor(
        'SFMT', True, 1, default_value='C',
        docstring='Graphic Type. This field shall contain a valid indicator of the '
                  'representation type of the graphic.')  # type: str
    SSTRUCT = _IntegerDescriptor(
        'SSTRUCT', True, 13, default_value=0,
        docstring='Reserved for Future Use.')  # type: int
    SDLVL = _IntegerDescriptor(
        'SDLVL', True, 3, default_value=1,
        docstring='Graphic Display Level. This field shall contain a valid value that indicates '
                  'the graphic display level of the graphic relative to other displayed file '
                  'components in a composite display. The valid values are :code:`1-999`. '
                  'The display level of each displayable file component (image or graphic) '
                  'within a file shall be unique.')  # type: int
    SALVL = _IntegerDescriptor(
        'SALVL', True, 3, default_value=0,
        docstring='Graphic Attachment Level. This field shall contain a valid value '
                  'that indicates the attachment level of the graphic. Valid values for '
                  'this field are 0 and the display level value of any other '
                  'image or graphic in the file.')  # type: int
    SLOC = _IntegerDescriptor(
        'SLOC', True, 10, default_value=0,
        docstring='Graphic Location. The graphics location is specified by providing the location '
                  'of the graphics origin point relative to the position (location of the CCS, image, '
                  'or graphic to which it is attached. This field shall contain the graphic location '
                  'offset from the `ILOC` or `SLOC` value of the CCS, image, or graphic to which the graphic '
                  'is attached or from the origin of the CCS when the graphic is unattached (`SALVL = 0`). '
                  'A row and column value of :code:`0` indicates no offset. Positive row and column values indicate '
                  'offsets down and to the right, while negative row and column values indicate '
                  'offsets up and to the left.')  # type: int
    SBND1 = _IntegerDescriptor(
        'SBND1', True, 10, default_value=0,
        docstring='First Graphic Bound Location. This field shall contain an ordered pair of '
                  'integers defining a location in Cartesian coordinates for use with CGM graphics. It is '
                  'the upper left corner of the bounding box for the CGM graphic.')  # type: int
    SCOLOR = _StringEnumDescriptor(
        'SCOLOR', True, 1, {'C', 'M'}, default_value='M',
        docstring='Graphic Color. If `SFMT = C`, this field shall contain a :code:`C` if the CGM contains any '
                  'color pieces or an :code:`M` if it is monochrome (i.e., black, '
                  'white, or levels of grey).')  # type: str
    SBND2 = _IntegerDescriptor(
        'SBND2', True, 10, default_value=0,
        docstring='Second Graphic Bound Location. This field shall contain an ordered pair of '
                  'integers defining a location in Cartesian coordinates for use with CGM graphics. '
                  'It is the lower right corner of the bounding box for the CGM graphic.')  # type: int
    SRES2 = _IntegerDescriptor(
        'SRES2', True, 2, default_value=0,
        docstring='Reserved for Future Use.')  # type: int
    UserHeader = _NITFElementDescriptor(
        'UserHeader', True, UserHeaderType, default_args={},
        docstring='User defined header.')  # type: UserHeaderType



from __future__ import print_function
import sys
import functools
from collections import defaultdict

import pkgutil
from importlib import import_module

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


print_func = print


def traverse_module_classification(package_name, results_dict):
    def evaluate(the_module, the_name):
        class_str = getattr(the_module, '__classification__', '__NO_CLASSIFICATION__')
        results_dict[class_str].append(the_name)

    module = import_module(package_name)

    if hasattr(module, '__path__'):
        for details in pkgutil.walk_packages(module.__path__, package_name+'.'):
            _, module_name, is_pkg = details
            if is_pkg:
                # don't evaluate the presence of class string for packages
                continue

            # noinspection PyBroadException
            sub_module = import_module(module_name)
            evaluate(sub_module, module_name)
    else:
        evaluate(module, package_name)


def check_classification(parent_package, results_dict=None):
    if results_dict is None:
        results_dict = defaultdict(list)
    traverse_module_classification(parent_package, results_dict)
    return results_dict


def log_package_classification(parent_package, dest=sys.stdout):
    global print_func
    print_func = functools.partial(print, file=dest)

    results_dict = check_classification(parent_package)
    for class_str in sorted(results_dict.keys()):
        print_func(class_str)
        for entry in results_dict[class_str]:
            print_func('\t', entry)


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description='Utility to create a report for displaying package __classification__ values',
        formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-p', '--package', default='sarpy',
                        help="package or module name, should be a subpackage of sarpy")
    parser.add_argument('-o', '--output', default='stdout',
                        help="'stdout', 'string', or an output file")
    args = parser.parse_args()

    if args.output == 'stdout':
        # Send output to stdout
        log_package_classification(args.package, dest=sys.stdout)
    else:
        # Send output to file
        with open(args.output, 'w') as f:
            log_package_classification(args.package, dest=f)
